import types
from collections import namedtuple
from io import StringIO
from textwrap import dedent

import numpy
import numpy.testing as nptest
import pandas
import pandas.testing as pdtest
import pytest
import statsmodels.api as sm
from scipy import stats

from wqio.tests import helpers
from wqio.utils import numutils


@pytest.mark.parametrize(
    ("value", "N", "tex", "forceint", "pval", "expected", "error"),
    [
        (1234.56, 3, False, False, False, "1,230", None),
        (1234.56, 4, False, False, False, "1,235", None),
        (1234.56, 8, False, False, False, "1,234.5600", None),
        (1234.56, 0, False, False, False, None, ValueError),
        (1234.56, -1, False, False, False, None, ValueError),
        (1234.56 * 1e5, 3, False, False, False, "1.23e+08", None),
        (1234.56 * 1e5, 3, True, False, False, r"$1.23 \times 10 ^ {8}$", None),
        (1234.56, 3, False, True, False, "1,235", None),
        (0.123456, 3, False, False, False, "0.123", None),
        (0.123456, 4, False, False, False, "0.1235", None),
        (0.123456, 8, False, False, False, "0.12345600", None),
        (0.123456, -1, False, False, False, None, ValueError),
        (0.123456, 0, False, False, False, None, ValueError),
        (0.123456 * 1e-6, 3, False, False, False, "1.23e-07", None),
        (0.123456 * 1e-6, 3, True, False, False, r"$1.23 \times 10 ^ {-7}$", None),
        (0.123456, 3, False, True, False, "0", None),
        (0.001, 1, False, False, True, "0.001", None),
        (0.0005, 3, False, False, True, "<0.001", None),
        (0.0005, 3, True, False, True, "$<0.001$", None),
        ("1234", 3, False, False, False, "1234", None),
        (numpy.nan, 3, False, False, False, "NA", None),
        (numpy.inf, 3, False, False, False, "NA", None),
        (numpy.inf * -1, 3, False, False, False, "NA", None),
    ],
)
def test_sig_figs(value, N, tex, forceint, pval, expected, error):
    with helpers.raises(error):
        result = numutils.sig_figs(value, N, tex=tex, forceint=forceint, pval=pval)
        assert result == expected


@pytest.mark.parametrize(
    ("num", "qual", "N", "expected"),
    [
        (12498.124, "<", 3, "<12,500"),
        (12.540, ">", 6, ">12.5400"),
        (12.540, "", 3, "12.5"),
        (0.00257, "=", 2, "=0.0026"),
    ],
)
def test_format_result(num, qual, N, expected):
    assert numutils.format_result(num, qual, N) == expected


@pytest.mark.parametrize(
    ("fxn", "pval", "expected", "error_to_raise"),
    [
        (numutils.process_p_vals, None, "NA", None),
        (numutils.process_p_vals, 0.0005, "<0.001", None),
        (numutils.process_p_vals, 0.005, "0.005", None),
        (numutils.process_p_vals, 0.001, "0.001", None),
        (numutils.process_p_vals, 0.0012, "0.001", None),
        (numutils.process_p_vals, 0.06, "0.060", None),
        (numutils.process_p_vals, 1.01, None, ValueError),
    ],
)
def test_process_p_vals(fxn, pval, expected, error_to_raise):
    with helpers.raises(error_to_raise):
        result = fxn(pval)
        assert result == expected


@pytest.mark.parametrize("as_emoji", [False, True])
@pytest.mark.parametrize(
    ("pval", "expected", "error_to_raise"),
    [
        (None, ("NA", r"ಠ_ಠ"), None),
        (0.005, ("maybe", r"¯\(ツ)/¯"), None),
        (0.03, ("maybe (weak)", r"¯\_(ツ)_/¯"), None),
        (0.06, ("maybe (very weak)", r"¯\__(ツ)__/¯"), None),
        (0.11, ("nope", r"(╯°□°)╯︵ ┻━┻"), None),
        (1.01, (None, None), ValueError),
    ],
)
def test_translate_p_vals(pval, expected, as_emoji, error_to_raise):
    with helpers.raises(error_to_raise):
        result = numutils.translate_p_vals(pval, as_emoji=as_emoji)
        assert result == expected[as_emoji]


def test_anderson_darling():
    data = helpers.getTestROSData()
    result = numutils.anderson_darling(data["res"])

    tolerance = 0.05
    known_anderson = (
        1.4392085,
        [0.527, 0.6, 0.719, 0.839, 0.998],
        [15.0, 10.0, 5.0, 2.5, 1.0],
        None,
        0.00080268,
    )

    # Ad stat
    assert abs(result[0] - known_anderson[0]) < 0.0001

    # critical values
    nptest.assert_allclose(result[1], known_anderson[1], rtol=tolerance)

    # significance level
    nptest.assert_allclose(result[2], known_anderson[2], rtol=tolerance)

    # p-value
    assert abs(result[4] - known_anderson[4]) < 0.0000001


@pytest.mark.parametrize("which", ["good", "bad"])
def test_processAndersonDarlingResults(which):
    fieldnames = ["statistic", "critical_values", "significance_level"]
    AndersonResult = namedtuple("AndersonResult", fieldnames)
    ARs = {
        "bad": AndersonResult(
            statistic=0.30194681312357829,
            critical_values=numpy.array([0.529, 0.602, 0.722, 0.842, 1.002]),
            significance_level=numpy.array([15.0, 10.0, 5.0, 2.5, 1.0]),
        ),
        "good": AndersonResult(
            statistic=numpy.inf,
            critical_values=numpy.array([0.907, 1.061, 1.32, 1.58, 1.926]),
            significance_level=numpy.array([15.0, 10.0, 5.0, 2.50, 1.0]),
        ),
    }

    expected = {"bad": "<85.0%", "good": "99.0%"}

    assert numutils.process_AD_result(ARs[which]) == expected[which]


@pytest.mark.parametrize(
    ("AD", "n_points", "expected"),
    [
        (0.302, 1, 0.0035899),
        (0.302, 5, 0.4155200),
        (0.302, 15, 0.5325205),
        (0.150, 105, 0.9616770),
    ],
)
def test__anderson_darling_p_vals(AD, n_points, expected):
    ad_result = (AD, None, None)
    p_val = numutils._anderson_darling_p_vals(ad_result, n_points)
    assert abs(p_val - expected) < 0.0001


@pytest.fixture
def units_norm_data():
    data_csv = StringIO(
        dedent(
            """\
        storm,param,station,units,conc
        1,"Lead, Total",inflow,ug/L,10
        2,"Lead, Total",inflow,mg/L,0.02
        3,"Lead, Total",inflow,g/L,0.00003
        4,"Lead, Total",inflow,ug/L,40
        1,"Lead, Total",outflow,mg/L,0.05
        2,"Lead, Total",outflow,ug/L,60
        3,"Lead, Total",outflow,ug/L,70
        4,"Lead, Total",outflow,g/L,0.00008
        1,"Cadmium, Total",inflow,ug/L,10
        2,"Cadmium, Total",inflow,mg/L,0.02
        3,"Cadmium, Total",inflow,g/L,0.00003
        4,"Cadmium, Total",inflow,ug/L,40
        1,"Cadmium, Total",outflow,mg/L,0.05
        2,"Cadmium, Total",outflow,ug/L,60
        3,"Cadmium, Total",outflow,ug/L,70
        4,"Cadmium, Total",outflow,g/L,0.00008
    """
        )
    )
    raw = pandas.read_csv(data_csv)

    known_csv = StringIO(
        dedent(
            """\
        storm,param,station,units,conc
        1,"Lead, Total",inflow,ug/L,10.
        2,"Lead, Total",inflow,ug/L,20.
        3,"Lead, Total",inflow,ug/L,30.
        4,"Lead, Total",inflow,ug/L,40.
        1,"Lead, Total",outflow,ug/L,50.
        2,"Lead, Total",outflow,ug/L,60.
        3,"Lead, Total",outflow,ug/L,70.
        4,"Lead, Total",outflow,ug/L,80.
        1,"Cadmium, Total",inflow,mg/L,0.010
        2,"Cadmium, Total",inflow,mg/L,0.020
        3,"Cadmium, Total",inflow,mg/L,0.030
        4,"Cadmium, Total",inflow,mg/L,0.040
        1,"Cadmium, Total",outflow,mg/L,0.050
        2,"Cadmium, Total",outflow,mg/L,0.060
        3,"Cadmium, Total",outflow,mg/L,0.070
        4,"Cadmium, Total",outflow,mg/L,0.080
    """
        )
    )
    expected = pandas.read_csv(known_csv)
    return raw, expected


def test_normalize_units(units_norm_data):
    unitsmap = {"ug/L": 1e-6, "mg/L": 1e-3, "g/L": 1e0}

    targetunits = {"Lead, Total": "ug/L", "Cadmium, Total": "mg/L"}
    raw, expected = units_norm_data
    result = numutils.normalize_units(
        raw, unitsmap, targetunits, paramcol="param", rescol="conc", unitcol="units"
    )
    pdtest.assert_frame_equal(result, expected)


def test_normalize_units_bad_targetunits(units_norm_data):
    unitsmap = {"ug/L": 1e-6, "mg/L": 1e-3, "g/L": 1e0}

    targetunits = {"Lead, Total": "ug/L"}
    raw, expected = units_norm_data
    with helpers.raises(ValueError):
        numutils.normalize_units(
            raw,
            unitsmap,
            targetunits,
            paramcol="param",
            rescol="conc",
            unitcol="units",
            napolicy="raise",
        )


def test_normalize_units_bad_normalization(units_norm_data):
    unitsmap = {"mg/L": 1e-3, "g/L": 1e0}

    targetunits = {"Lead, Total": "ug/L", "Cadmium, Total": "mg/L"}
    raw, expected = units_norm_data
    with helpers.raises(ValueError):
        numutils.normalize_units(
            raw,
            unitsmap,
            targetunits,
            paramcol="param",
            rescol="conc",
            unitcol="units",
            napolicy="raise",
        )


def test_normalize_units_bad_conversion(units_norm_data):
    unitsmap = {"ug/L": 1e-6, "mg/L": 1e-3, "g/L": 1e0}

    targetunits = {"Lead, Total": "ng/L", "Cadmium, Total": "mg/L"}
    raw, expected = units_norm_data
    with helpers.raises(ValueError):
        numutils.normalize_units(
            raw,
            unitsmap,
            targetunits,
            paramcol="param",
            rescol="conc",
            unitcol="units",
            napolicy="raise",
        )


@pytest.mark.parametrize(
    ("pH", "expected", "error"),
    [
        (4, 0.10072764682551091, None),
        (14.1, None, ValueError),
        (-0.1, None, ValueError),
    ],
)
def test_test_pH2concentration(pH, expected, error):
    with helpers.raises(error):
        assert abs(numutils.pH_to_concentration(pH) - expected) < 0.0001


@helpers.seed
@pytest.mark.parametrize("error", [None, ValueError])
def test_compute_theilslope_default(error):
    with helpers.raises(error):
        y = helpers.getTestROSData()["res"].values
        x = numpy.arange(len(y) - 1) if error else None
        assert tuple(numutils.compute_theilslope(y, x)) == stats.mstats.theilslopes(y, x)


@pytest.fixture
def fit_data():
    data = {
        "data": numpy.array(
            [
                2.000,
                4.000,
                4.620,
                5.000,
                5.000,
                5.500,
                5.570,
                5.660,
                5.750,
                5.860,
                6.650,
                6.780,
                6.790,
                7.500,
                7.500,
                7.500,
                8.630,
                8.710,
                8.990,
                9.500,
                9.500,
                9.850,
                10.82,
                11.00,
                11.25,
                11.25,
                12.20,
                14.92,
                16.77,
                17.81,
                19.16,
                19.19,
                19.64,
                20.18,
                22.97,
            ]
        ),
        "zscores": numpy.array(
            [
                -2.06188401,
                -1.66883254,
                -1.43353970,
                -1.25837339,
                -1.11509471,
                -0.99166098,
                -0.88174260,
                -0.78156696,
                -0.68868392,
                -0.60139747,
                -0.51847288,
                -0.43897250,
                -0.36215721,
                -0.28742406,
                -0.21426459,
                -0.14223572,
                -0.07093824,
                0.00000000,
                0.07093824,
                0.14223572,
                0.21426459,
                0.28742406,
                0.36215721,
                0.43897250,
                0.51847288,
                0.60139747,
                0.68868392,
                0.78156696,
                0.88174260,
                0.99166098,
                1.11509471,
                1.25837339,
                1.43353970,
                1.66883254,
                2.06188401,
            ]
        ),
        "y": numpy.array(
            [
                0.07323274,
                0.12319301,
                0.16771455,
                0.17796950,
                0.21840761,
                0.25757016,
                0.27402650,
                0.40868106,
                0.44872637,
                0.53673530,
                0.55169933,
                0.56211726,
                0.62375442,
                0.66631353,
                0.68454978,
                0.72137134,
                0.87602096,
                0.94651962,
                1.01927875,
                1.06040448,
                1.07966792,
                1.17969506,
                1.21132273,
                1.30751428,
                1.45371899,
                1.76381932,
                1.98832275,
                2.09275652,
                2.66552831,
                2.86453334,
                3.23039631,
                4.23953492,
                4.25892247,
                4.58347660,
                6.53100725,
            ]
        ),
    }
    z2, _y = stats.probplot(data["y"], fit=False)
    data["probs"] = stats.norm.cdf(data["zscores"]) * 100.0
    data["p2"] = stats.norm.cdf(z2) * 100
    return data


@pytest.mark.parametrize(
    ("fitlogs", "fitprobs", "error"),
    [
        (None, None, None),
        (None, None, None),
        ("y", None, None),
        (None, "y", None),
        ("x", None, None),
        ("both", None, None),
        ("x", "y", None),
        (None, "x", None),
        ("y", "x", None),
        (None, "both", None),
        ("junk", None, ValueError),
        (None, "junk", ValueError),
    ],
)
def test_fit_line(fit_data, fitlogs, fitprobs, error):
    xy = {
        (None, None): (fit_data["zscores"], fit_data["data"]),
        ("y", None): (fit_data["zscores"], fit_data["data"]),
        (None, "y"): (fit_data["data"], fit_data["probs"]),
        ("x", None): (fit_data["data"], fit_data["zscores"]),
        ("both", None): (fit_data["data"], fit_data["y"]),
        ("x", "y"): (fit_data["data"], fit_data["probs"]),
        (None, "x"): (fit_data["probs"], fit_data["data"]),
        ("y", "x"): (fit_data["probs"], fit_data["data"]),
        (None, "both"): (fit_data["probs"], fit_data["p2"]),
        ("junk", None): (fit_data["zscores"], fit_data["data"]),
        (None, "junk"): (fit_data["zscores"], fit_data["data"]),
    }

    expected = {
        (None, None): numpy.array([-0.89650596, 21.12622025]),
        ("y", None): numpy.array([2.80190754, 27.64958934]),
        (None, "y"): numpy.array([8.48666156, 98.51899616]),
        ("x", None): numpy.array([-2.57620461, 1.66767934]),
        ("both", None): numpy.array([0.04681540, 5.73261406]),
        ("x", "y"): numpy.array([0.49945757, 95.23103009]),
        (None, "x"): numpy.array([-0.89650596, 21.12622025]),
        ("y", "x"): numpy.array([2.80190754, 27.64958934]),
        (None, "both"): numpy.array([1.96093902, 98.03906098]),
        ("junk", None): None,
        (None, "junk"): None,
    }

    with helpers.raises(error):
        x_, y_, res = numutils.fit_line(
            xy[(fitlogs, fitprobs)][0],
            xy[(fitlogs, fitprobs)][1],
            fitlogs=fitlogs,
            fitprobs=fitprobs,
        )
        nptest.assert_array_almost_equal(y_, expected[(fitlogs, fitprobs)])
        assert isinstance(res, sm.regression.linear_model.RegressionResultsWrapper)


def test_fit_line_through_origin(fit_data):
    x, y = fit_data["zscores"], fit_data["data"]
    x_, y_, res = numutils.fit_line(x, y, through_origin=True)
    assert res.params[0] == 0


def test_fit_line_with_xhat(fit_data):
    x, y = fit_data["zscores"], fit_data["data"]
    x_, y_, res = numutils.fit_line(x, y, xhat=[-2, -1, 0, 1, 2])
    expected = [-0.566018, 4.774419, 10.114857, 15.455295, 20.795733]
    nptest.assert_array_almost_equal(y_, expected)


@pytest.mark.parametrize(
    ("oneway", "expected"), [(True, [False, False, True]), (False, [False, True, True])]
)
def test_checkIntervalOverlap_single(oneway, expected):
    result = [
        numutils.check_interval_overlap([1, 2], [3, 4], oneway=oneway),
        numutils.check_interval_overlap([1, 4], [2, 3], oneway=oneway),
        numutils.check_interval_overlap([1, 3], [2, 4], oneway=oneway),
    ]
    assert result == expected


@pytest.mark.parametrize(("oneway", "expected"), [(True, [0, 0, 1]), (False, [0, 1, 1])])
def test_checkIntervalOverlap(oneway, expected):
    x = numpy.array([[1, 2], [1, 4], [1, 3]])
    y = numpy.array([[3, 4], [2, 3], [2, 4]])
    nptest.assert_array_equal(
        numutils.check_interval_overlap(x, y, oneway=oneway, axis=1),
        numpy.array(expected, dtype=bool),
    )


@pytest.mark.parametrize(
    ("opts", "expected_key"),
    [
        (dict(), "no-op"),
        (dict(A=0.05), "one-col"),
        (dict(A=0.05, C=0.10), "two-col"),
        (dict(A=0.20, B=0.20, C=0.10), "three-col"),
        (dict(A=(0.05, 0.20), B=0.20, C=0.10), "tuple-limit"),
    ],
)
def test_winsorize_dataframe(opts, expected_key):
    x = numpy.array([0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20])

    w_05 = numpy.array([1, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 19])

    w_10 = numpy.array([2, 2, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 18, 18])

    w_20 = numpy.array([4, 4, 4, 4, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 16, 16, 16, 16])

    w_05_20 = numpy.array(
        [1, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 16, 16, 16, 16]
    )

    df = pandas.DataFrame({"A": x, "B": x, "C": x})
    expected = {
        "no-op": df.copy(),
        "one-col": pandas.DataFrame({"A": w_05, "B": x, "C": x}),
        "two-col": pandas.DataFrame({"A": w_05, "B": x, "C": w_10}),
        "three-col": pandas.DataFrame({"A": w_20, "B": w_20, "C": w_10}),
        "tuple-limit": pandas.DataFrame({"A": w_05_20, "B": w_20, "C": w_10}),
    }

    result = numutils.winsorize_dataframe(df, **opts)
    assert result.eq(expected[expected_key]).all().all()


def test__comp_stat_generator():
    df = helpers.make_dc_data().reset_index()
    gen = numutils._comp_stat_generator(df, ["param", "bmp"], "loc", "res", helpers.comp_statfxn)
    assert isinstance(gen, types.GeneratorType)
    result = pandas.DataFrame(gen)

    expected = {
        "bmp": ["1", "1", "1", "1", "1", "7", "7", "7", "7", "7"],
        "param": ["A", "A", "A", "A", "A", "H", "H", "H", "H", "H"],
        "loc_1": [
            "Inflow",
            "Inflow",
            "Outflow",
            "Outflow",
            "Reference",
            "Inflow",
            "Outflow",
            "Outflow",
            "Reference",
            "Reference",
        ],
        "loc_2": [
            "Outflow",
            "Reference",
            "Inflow",
            "Reference",
            "Inflow",
            "Reference",
            "Inflow",
            "Reference",
            "Inflow",
            "Outflow",
        ],
        "stat": [
            33.100128,
            34.228203,
            18.318497,
            21.231383,
            9.500766,
            2.990900,
            7.498225,
            7.757954,
            3.067299,
            3.290471,
        ],
        "pvalue": [
            8.275032,
            8.557050,
            4.579624,
            5.307845,
            2.375191,
            0.747725,
            1.874556,
            1.939488,
            0.766824,
            0.822617,
        ],
    }
    pdtest.assert_frame_equal(
        pandas.DataFrame(expected, index=[0, 1, 2, 3, 4, 331, 332, 333, 334, 335]).sort_index(
            axis="columns"
        ),
        pandas.concat([result.head(), result.tail()]).sort_index(axis="columns"),
    )


def test__comp_stat_generator_single_group_col():
    df = helpers.make_dc_data().reset_index()
    gen = numutils._comp_stat_generator(df, "param", "loc", "res", helpers.comp_statfxn)
    assert isinstance(gen, types.GeneratorType)
    result = pandas.DataFrame(gen)

    expected = {
        "stat": [
            35.78880963,
            36.032519607,
            24.07328047,
            24.151276229,
            13.107839675,
            16.69316696,
            14.336154663,
            14.33615466,
            11.298920232,
            11.298920232,
        ],
        "loc_1": [
            "Inflow",
            "Inflow",
            "Outflow",
            "Outflow",
            "Reference",
            "Inflow",
            "Outflow",
            "Outflow",
            "Reference",
            "Reference",
        ],
        "loc_2": [
            "Outflow",
            "Reference",
            "Inflow",
            "Reference",
            "Inflow",
            "Reference",
            "Inflow",
            "Reference",
            "Inflow",
            "Outflow",
        ],
        "param": ["A", "A", "A", "A", "A", "H", "H", "H", "H", "H"],
        "pvalue": [
            8.94720240,
            9.008129901,
            6.018320119,
            6.037819057,
            3.276959918,
            4.17329174,
            3.584038665,
            3.584038665,
            2.824730058,
            2.824730058,
        ],
    }
    pdtest.assert_frame_equal(
        pandas.DataFrame(expected, index=[0, 1, 2, 3, 4, 43, 44, 45, 46, 47]).sort_index(
            axis="columns"
        ),
        pandas.concat([result.head(), result.tail()]).sort_index(axis="columns"),
    )


def test__paired_stat_generator():
    df = helpers.make_dc_data_complex().unstack(level="loc")
    gen = numutils._paired_stat_generator(df, ["param"], "loc", "res", helpers.comp_statfxn)
    assert isinstance(gen, types.GeneratorType)
    result = pandas.DataFrame(gen).sort_index(axis="columns")

    expected = pandas.DataFrame(
        {
            "loc_1": [
                "Inflow",
                "Inflow",
                "Outflow",
                "Outflow",
                "Reference",
                "Inflow",
                "Outflow",
                "Outflow",
                "Reference",
                "Reference",
            ],
            "loc_2": [
                "Outflow",
                "Reference",
                "Inflow",
                "Reference",
                "Inflow",
                "Reference",
                "Inflow",
                "Reference",
                "Inflow",
                "Outflow",
            ],
            "param": ["A", "A", "A", "A", "A", "H", "H", "H", "H", "H"],
            "pvalue": [
                2.688485,
                3.406661,
                9.084853,
                9.084853,
                5.243408,
                9.399253,
                20.234093,
                20.23409,
                2.801076,
                2.801075,
            ],
            "stat": [
                10.753940,
                13.626645,
                36.339414,
                36.339414,
                20.973631,
                37.597010,
                80.936373,
                80.936373,
                11.204302,
                11.204302,
            ],
        },
        index=[0, 1, 2, 3, 4, 43, 44, 45, 46, 47],
    ).sort_index(axis="columns")

    pdtest.assert_frame_equal(expected, pandas.concat([result.head(), result.tail()]))


@helpers.seed
def test_remove_outliers():
    expected_shape = (35,)
    x = numpy.random.normal(0, 4, size=37)

    assert numutils.remove_outliers(x).shape == expected_shape



import pandas
import seaborn
from matplotlib import pyplot
from pandas.plotting import register_matplotlib_converters

from wqio import utils

register_matplotlib_converters()


class Parameter:
    def __init__(self, name, units, usingTex=False):
        """Class representing a single analytical parameter (pollutant).

        (Input) Parameters
        ------------------
        name : string
            Name of the parameter.
        units : string
            Units of measure for the parameter.
        usingTex : bool, optional (default = False)
            If True, all returned values will be optimized for inclusion
            in LaTeX documents.

        """

        self._name = name
        self._units = units
        self._usingTex = usingTex

    @property
    def name(self):
        return self._name

    @name.setter
    def name(self, value):
        self._name = value

    @property
    def units(self):
        return self._units

    @units.setter
    def units(self, value):
        self._units = value

    @property
    def usingTex(self):
        return self._usingTex

    @usingTex.setter
    def usingTex(self, value):
        if value in (True, False):
            self._usingTex = value
        else:
            raise ValueError("`usingTex` must be of type `bool`")

    def paramunit(self, usecomma=False):
        """Creates a string representation of the parameter and units.

        Parameters
        ----------
        usecomma : bool, optional (default = False)
            Toggles the format of the returned string attribute. If True
            the returned format is "<parameter>, <unit>". Otherwise the
            format is "<parameter> (<unit>)".
        """

        paramunit = "{0}, {1}" if usecomma else "{0} ({1})"

        n = self.name
        u = self.units

        return paramunit.format(n, u)

    def __repr__(self):
        return f"<wqio Parameter object> ({self.paramunit(usecomma=False)})"

    def __str__(self):
        return f"<wqio Parameter object> ({self.paramunit(usecomma=False)})"


class SampleMixin:
    def __init__(
        self,
        dataframe,
        starttime,
        samplefreq=None,
        endtime=None,
        storm=None,
        rescol="res",
        qualcol="qual",
        dlcol="DL",
        unitscol="units",
    ):
        self._wqdata = dataframe
        self._startime = pandas.Timestamp(starttime)
        self._endtime = pandas.Timestamp(endtime)
        self._samplefreq = samplefreq
        self._sample_ts = None
        self._label = None
        self._marker = None
        self._markersize = None
        self._linestyle = None
        self._yfactor = None
        self._season = utils.get_season(self.starttime)
        self.storm = storm

    @property
    def season(self):
        return self._season

    @season.setter
    def season(self, value):
        self._season = value

    @property
    def wqdata(self):
        return self._wqdata

    @wqdata.setter
    def wqdata(self, value):
        self._wqdata = value

    @property
    def starttime(self):
        return self._startime

    @starttime.setter
    def starttime(self, value):
        self._startime = value

    @property
    def endtime(self):
        if self._endtime is None:
            self._endtime = self._startime
        return self._endtime

    @endtime.setter
    def endtime(self, value):
        self._endtime = value

    @property
    def samplefreq(self):
        return self._samplefreq

    @samplefreq.setter
    def samplefreq(self, value):
        self._samplefreq = value

    @property
    def linestyle(self):
        if self._linestyle is None:
            self._linestyle = "none"
        return self._linestyle

    @linestyle.setter
    def linestyle(self, value):
        self._linestyle = value

    @property
    def markersize(self):
        if self._markersize is None:
            self._markersize = 4
        return self._markersize

    @markersize.setter
    def markersize(self, value):
        self._markersize = value

    @property
    def yfactor(self):
        if self._yfactor is None:
            self._yfactor = 0.25
        return self._yfactor

    @yfactor.setter
    def yfactor(self, value):
        self._yfactor = value

    def plot_ts(self, ax, isFocus=True, asrug=False):
        if self.sample_ts is not None:
            alpha = 0.75 if isFocus else 0.35

        ymax = ax.get_ylim()[-1]
        yposition = [self.yfactor * ymax] * len(self.sample_ts)

        timeseries = pandas.Series(yposition, index=self.sample_ts)

        if asrug:
            seaborn.rugplot(self.sample_ts, ax=ax, color="black", alpha=alpha, mew=0.75)
            line = pyplot.Line2D(
                [0, 0],
                [0, 0],
                marker="|",
                mew=0.75,
                color="black",
                alpha=alpha,
                linestyle="none",
            )

        else:
            timeseries.plot(
                ax=ax,
                marker=self.marker,
                markersize=4,
                linestyle=self.linestyle,
                color="Black",
                zorder=10,
                label="_nolegend",
                alpha=alpha,
                mew=0.75,
            )
            line = pyplot.Line2D(
                [0, 0],
                [0, 0],
                marker=self.marker,
                mew=0.75,
                color="black",
                alpha=alpha,
                linestyle="none",
            )

        return line


class CompositeSample(SampleMixin):
    """Class for composite samples"""

    @property
    def label(self):
        if self._label is None:
            self._label = "Composite Sample"
        return self._label

    @label.setter
    def label(self, value):
        self._label = value

    @property
    def marker(self):
        if self._marker is None:
            self._marker = "x"
        return self._marker

    @marker.setter
    def marker(self, value):
        self._marker = value

    @property
    def sample_ts(self):
        if self.starttime is not None and self.endtime is not None:
            _sampfreq = self.samplefreq or self.endtime - self.starttime
            self._sample_ts = pandas.date_range(
                start=self.starttime, end=self.endtime, freq=_sampfreq
            )
        return self._sample_ts


class GrabSample(SampleMixin):
    """Class for grab (discrete) samples"""

    @property
    def label(self):
        if self._label is None:
            self._label = "Grab Sample"
        return self._label

    @label.setter
    def label(self, value):
        self._label = value

    @property
    def marker(self):
        if self._marker is None:
            self._marker = "+"
        return self._marker

    @marker.setter
    def marker(self, value):
        self._marker = value

    @property
    def sample_ts(self):
        if self._sample_ts is None and self.starttime is not None:
            if self.endtime is None:
                self._sample_ts = pandas.DatetimeIndex(data=[self.starttime])
            else:
                self._sample_ts = pandas.date_range(
                    start=self.starttime,
                    end=self.endtime,
                    freq=self.endtime - self.starttime,
                )
        return self._sample_ts


_basic_doc = """ {} water quality sample

Container to hold water quality results from many different pollutants
collected at a single point in time.

Parameters
----------
dataframe : pandas.DataFrame
    The water quality data.
starttime : datetime-like
    The date and time at which the sample collection started.
samplefreq : string, optional
    A valid pandas timeoffset string specifying the frequency with which
    sample aliquots are collected.
endtime : datetime-like, optional
    The date and time at which sample collection ended.
storm : wqio.Storm object, optional
    A Storm object (or a subclass or Storm) that triggered sample
    collection.
rescol, qualcol, dlcol, unitscol : string, optional
    Strings that define the column labels for the results, qualifiers,
    detection limits, and units if measure, respectively.

"""


SampleMixin.__doc__ = _basic_doc.format("Basic")
CompositeSample.__doc__ = _basic_doc.format("Composite")
GrabSample.__doc__ = _basic_doc.format("Grab")



import itertools
from collections import namedtuple
from textwrap import dedent

import numpy
import statsmodels.api as sm
from probscale.algo import _estimate_from_fit
from scipy import stats

from wqio import validate
from wqio.utils import misc

TheilStats = namedtuple("TheilStats", ("slope", "intercept", "low_slope", "high_slope"))


def sig_figs(x, n, expthresh=5, tex=False, pval=False, forceint=False):
    """Formats a number with the correct number of sig figs.

    Parameters
    ----------
    x : int or float
        The number you want to format.
    n : int
        The number of sig figs it should have.
    expthresh : int, optional (default = 5)
        The absolute value of the order of magnitude at which numbers
        are formatted in exponential notation.
    tex : bool, optional (default is False)
        Toggles the scientific formatting of the number either for
        terminal output (False) or LaTeX documents (True).
    pval : bool, optional (default is False)
        Useful for formatting p-values from hypothesis tests. When True
        and x < 0.001, will return "<0.001".
    forceint : bool, optional (default is False)
        If true, simply returns int(x)

    Returns
    -------
    formatted : str
        The formatted number as a string

    Examples
    --------
    >>> print(sig_figs(1247.15, 3))
    1,250

    >>> print(sig_figs(1247.15, 7))
    1,247.150

    """
    # return a string value unaltered
    if isinstance(x, str):
        out = x

    # check on the number provided
    elif x is not None and not numpy.isinf(x) and not numpy.isnan(x):
        # check on the sig_figs
        if n < 1:
            raise ValueError("number of sig figs must be greater than zero!")

        elif pval and x < 0.001:
            out = "<0.001"
            if tex:
                out = f"${out}$"

        elif forceint:
            out = f"{x:,.0f}"

        # logic to do all of the rounding
        elif x != 0.0:
            order = numpy.floor(numpy.log10(numpy.abs(x)))

            if -1.0 * expthresh <= order <= expthresh:
                decimal_places = int(n - 1 - order)

                if decimal_places <= 0:
                    out = f"{round(x, decimal_places):,.0f}"

                else:
                    fmt = "{0:,.%df}" % decimal_places
                    out = fmt.format(x)

            else:
                decimal_places = n - 1
                if tex:
                    fmt = r"$%%0.%df \times 10 ^ {%d}$" % (decimal_places, order)
                    out = fmt % round(x / 10**order, decimal_places)
                else:
                    fmt = "{0:.%de}" % decimal_places
                    out = fmt.format(x)

        else:
            out = str(round(x, n))

    # with NAs and INFs, just return 'NA'
    else:
        out = "NA"

    return out


def format_result(result, qualifier, sigfigs=3):
    """Formats a results with its qualifier

    Parameters
    ----------
    results : float
        The concentration or particulate strength
    qualifier : string
        The result's qualifier
    sigfigs : int
        The number of significant digits to which `result` should be
        formatted

    Returns
    -------
    formatted : string

    Example
    -------
    >>> format_result(1.23, '<', sigfigs=4)
    '<1.230'

    """

    return f"{qualifier}{sig_figs(result, sigfigs)}"


def process_p_vals(pval):
    """Processes p-values into nice strings to reporting. When the
    p-values are less than 0.001, "<0.001" is returned. Otherwise, a
    string with three decimal places is returned.

    Parameters
    ----------
    pval : float
        The p-value to be formatted. Must be between 0 and 1. Otherwise
        an error is raised.

    Returns
    -------
    processed : string

    """
    if pval is None:
        out = "NA"
    elif 0 < pval < 0.001:
        out = format_result(0.001, "<", sigfigs=1)
    elif pval > 1 or pval < 0:
        raise ValueError(f"p-values must be between 0 and 1 (not {pval})")
    else:
        out = "%0.3f" % pval

    return out


def translate_p_vals(pval, as_emoji=True):
    """Translates ambiguous p-values into more meaningful emoji.

    Parameters
    ----------
    pval : float
    as_emoji : bool (default = True)
        Represent the p-vals as emoji. Otherwise, words will be used.

    Returns
    -------
    interpreted : string

    """

    if pval is None:
        interpreted = r"ಠ_ಠ" if as_emoji else "NA"
    elif pval <= 0.01:
        interpreted = r"¯\(ツ)/¯" if as_emoji else "maybe"
    elif 0.01 < pval <= 0.05:
        interpreted = r"¯\_(ツ)_/¯" if as_emoji else "maybe (weak)"
    elif 0.05 < pval <= 0.1:
        interpreted = r"¯\__(ツ)__/¯" if as_emoji else "maybe (very weak)"
    elif 0.1 < pval <= 1:
        interpreted = r"(╯°□°)╯︵ ┻━┻" if as_emoji else "nope"
    else:
        raise ValueError(f"p-values must be between 0 and 1 (not {pval})")

    return interpreted


def anderson_darling(data):
    """
    Compute the Anderson-Darling Statistic and p-value

    Parmeters
    ---------
    data : array-like

    Returns
    -------
    ADResult : namedtuple
        A collection of results. Keys are:

          * ``statistic``
          * ``critical_values``
          * ``significance_level``
          * ``pvalue``


    See Also
    --------
    scipy.stats.anderson

    """
    AD = stats.anderson(data)
    # field_names = list(AD._fields) + ["pvalue"]

    p = _anderson_darling_p_vals(AD, len(data))
    values = AD._asdict()
    values["pvalue"] = p

    ADResult = namedtuple("ADResult", values)
    return ADResult(**values)


def process_AD_result(ad_result):
    """Return a nice string of Anderson-Darling test results

    Parameters
    ----------
    ad_result : tuple or namedtuple
        The packed output from scipy.stats.anderson

    Returns
    -------
    result : str
        A string representation of the confidence in the result.

    """

    AD, crit, sig = ad_result
    try:
        ci = 100 - sig[crit < AD][-1]
        return f"{ci:0.1f}%"
    except IndexError:
        ci = 100 - sig[0]
        return f"<{ci:0.1f}%"


def _anderson_darling_p_vals(ad_results, n_points):
    """
    Computes the p-value of the Anderson-Darling Result

    Parameters
    ----------
    ad_result : tuple or namedtuple
        The packed output from scipy.stats.anderson
    n_points : int
        The number of points in the sample.

    Returns
    -------
    p : float

    References
    ----------
    R.B. D'Augostino and M.A. Stephens, Eds., 1986, Goodness-of-Fit
    Techniques, Marcel Dekker

    """

    AD, crit, sig = ad_results
    AD_star = AD * (1 + 0.75 / n_points + 2.25 / n_points**2)
    if AD_star >= 0.6:
        p = numpy.exp(1.2397 - (5.709 * AD_star) + (0.0186 * AD_star**2))
    elif 0.34 <= AD_star < 0.6:
        p = numpy.exp(0.9177 - (4.279 * AD_star) - (1.38 * AD_star**2))
    elif 0.2 < AD_star < 0.34:
        p = 1 - numpy.exp(-8.318 + (42.796 * AD_star) - (59.938 * AD_star**2))
    else:
        p = 1 - numpy.exp(-13.436 + (101.14 * AD_star) - (223.73 * AD_star**2))

    return p


def normalize_units(
    df,
    unitsmap,
    targetunit,
    paramcol="parameter",
    rescol="res",
    unitcol="units",
    napolicy="ignore",
):
    """
    Normalize units of measure in a dataframe.

    Parameters
    ----------
    df : pandas.DataFrame
        Dataframe contained results and units of measure data.
    unitsmap : dictionary
        Dictionary where keys are the units present in the df and
        values are the conversion factors required to standardize
        results to a common unit.
    targetunit : string or dict
        The desired final units of measure. Must be present in
        ``unitsmap``. If a string, all rows in df will be
        assigned ``targetunit``. If a dictionary, the units will be
        mapped from the dictionary using the values of ``paramcol`` as
        keys.
    paramcol, rescol, unitcol : string, optional
        Labels for the parameter, results, and units columns in the
        df.
    napolicy : string, optional
        Determines how mull/missing values are stored. By default,
        this is set to "ignore". Use "raise" to throw a ``ValueError``
        if when unit conversion or target units data cannot be found.

    Returns
    -------
    normalized : pandas.DataFrame
        Dataframe with normalized units of measure.

    """

    # determine the preferred units in the wqdata
    target = df[paramcol].map(targetunit)

    # factors to normialize to standard units
    normalization = df[unitcol].map(unitsmap)

    # factor to convert to preferred units
    conversion = target.map(unitsmap)

    if napolicy == "raise":
        msg = ""
        if target.isnull().any():
            nulls = df[target.isnull()][paramcol].unique()
            msg += f"Some target units could not be mapped to the {paramcol} column ({nulls})\n"

        if normalization.isnull().any():
            nulls = df[normalization.isnull()][unitcol].unique()
            msg += "Some normalization factors could not be mapped to the {} column ({})\n".format(
                unitcol, nulls
            )

        if conversion.isnull().any():
            nulls = target[conversion.isnull()]
            msg += f"Some conversion factors could not be mapped to the target units ({nulls})"

        if len(msg) > 0:
            raise ValueError(msg)

    # convert results
    normalized = df.assign(**{rescol: df[rescol] * normalization / conversion, unitcol: target})

    return normalized


def pH_to_concentration(pH, *args):
    """Converts pH values to proton concentrations in mg/L

    Parameters
    ----------
    pH : int or float
        Number between 0 and 14 representing the pH

    Returns
    -------
    conc : float
        Concentration of protons in mg/L

    """

    # check that we recieved a valid input:
    if pH < 0 or pH > 14:
        raise ValueError("pH = %f but must be between 0 and 14" % pH)

    # avogadro's number (items/mole)
    avogadro = 6.0221413e23

    # mass of a proton (kg)
    proton_mass = 1.672621777e-27

    # grams per kilogram
    kg2g = 1000

    # milligrams per gram
    g2mg = 1000

    return 10 ** (-1 * pH) * avogadro * proton_mass * kg2g * g2mg


def compute_theilslope(y, x=None, alpha=0.95, percentile=50):
    f""" Adapted from stats.mstats.theilslopes so that we can tweak the
    `percentile` parameter.
    https://goo.gl/nxPF54
    {dedent(stats.mstats.theilslopes.__doc__)}
    """

    # We copy both x and y so we can use _find_repeats.
    y = numpy.array(y).flatten()
    if x is None:
        x = numpy.arange(len(y), dtype=float)
    else:
        x = numpy.array(x, dtype=float).flatten()
        if len(x) != len(y):
            raise ValueError(f"Incompatible lengths ({len(y)} != {len(x)})")

    # Compute sorted slopes only when deltax > 0
    deltax = x[:, numpy.newaxis] - x
    deltay = y[:, numpy.newaxis] - y
    slopes = deltay[deltax > 0] / deltax[deltax > 0]
    slopes.sort()
    outslope = numpy.percentile(slopes, percentile)
    outinter = numpy.percentile(y, percentile) - outslope * numpy.median(x)
    # Now compute confidence intervals
    if alpha > 0.5:
        alpha = 1.0 - alpha

    z = stats.distributions.norm.ppf(alpha / 2.0)

    # This implements (2.6) from Sen (1968)
    _, nxreps = stats.mstats.find_repeats(x)
    _, nyreps = stats.mstats.find_repeats(y)
    nt = len(slopes)  # N in Sen (1968)
    ny = len(y)  # n in Sen (1968)

    # Equation 2.6 in Sen (1968):
    sigsq = (
        1
        / 18.0
        * (
            ny * (ny - 1) * (2 * ny + 5)
            - numpy.sum([k * (k - 1) * (2 * k + 5) for k in nxreps])
            - numpy.sum([k * (k - 1) * (2 * k + 5) for k in nyreps])
        )
    )

    # Find the confidence interval indices in `slopes`
    sigma = numpy.sqrt(sigsq)
    Ru = min(int(numpy.round((nt - z * sigma) / 2.0)), len(slopes) - 1)
    Rl = max(int(numpy.round((nt + z * sigma) / 2.0)) - 1, 0)
    delta = slopes[[Rl, Ru]]

    return TheilStats(outslope, outinter, delta[0], delta[1])


def fit_line(x, y, xhat=None, fitprobs=None, fitlogs=None, dist=None, through_origin=False):
    """Fits a line to x-y data in various forms (raw, log, prob scales)

    Parameters
    ----------
    x, y : array-like
        Independent and dependent data, respectively.
    xhat : array-like or None, optional
        The values at which yhat should should be estimated. If
        not provided, falls back to the sorted values of ``x``.
    fitprobs, fitlogs : str, options.
        Defines how data should be transformed. Valid values are
        'x', 'y', or 'both'. If using ``fitprobs``, variables should
        be expressed as a percentage, i.e.,
        Probablility transform = lambda x: ``dist``.ppf(x / 100.).
        Log transform = lambda x: numpy.log(x).
        Take care to not pass the same value to both ``fitlogs`` and
        ``figprobs`` as both transforms will be applied.
    dist : scipy.stats distribution or None, optional
        A fully-spec'd scipy.stats distribution such that ``dist.ppf``
        can be called. If not provided, defaults to scipt.stats.norm.
    through_origin : bool, optional (default = False)
        If True, the y-intercept of the best-fit line will be
        constrained to 0 during the regression.

    Returns
    -------
    xhat, yhat : numpy arrays
        Linear model estimates of ``x`` and ``y``.
    results : a statmodels result object
        The object returned by statsmodels.OLS.fit()

    """

    fitprobs = validate.fit_arguments(fitprobs, "fitprobs")
    fitlogs = validate.fit_arguments(fitlogs, "fitlogs")

    x = numpy.asarray(x)
    y = numpy.asarray(y)

    if xhat is None:
        xhat = numpy.array([numpy.min(x), numpy.max(x)])

    if dist is None:
        dist = stats.norm

    if fitprobs in ["x", "both"]:
        x = dist.ppf(x / 100.0)
        xhat = dist.ppf(numpy.array(xhat) / 100.0)

    if fitprobs in ["y", "both"]:
        y = dist.ppf(y / 100.0)

    if fitlogs in ["x", "both"]:
        x = numpy.log(x)

    if fitlogs in ["y", "both"]:
        y = numpy.log(y)

    x = sm.add_constant(x)

    # fix the y-intercept at 0 if requested
    # note that this is a nuance of the statsmodels package.
    # `x[:, 0] = 5` doesn't fix the intercept at 5, for instance.
    if through_origin:
        x[:, 0] = 0

    results = sm.OLS(y, x).fit()
    yhat = _estimate_from_fit(
        xhat,
        results.params[1],
        results.params[0],
        xlog=fitlogs in ["x", "both"],
        ylog=fitlogs in ["y", "both"],
    )

    if fitprobs in ["y", "both"]:
        yhat = 100.0 * dist.cdf(yhat)
    if fitprobs in ["x", "both"]:
        xhat = 100.0 * dist.cdf(xhat)

    return xhat, yhat, results


def check_interval_overlap(interval1, interval2, oneway=False, axis=None):
    """Checks if two numeric intervals overlaps.

    Parameters
    ----------
    interval1, interval2 : array like
        len = 2 sequences to compare
    oneway : bool (default = False)
        If True, only checks that ``interval1`` falls at least partially
        inside ``interval2``, but not the other way around

    Returns
    -------
    overlap : bool
        True if an overlap exists. Otherwise, False.

    """

    first_check = numpy.bitwise_or(
        numpy.bitwise_and(
            numpy.min(interval2, axis=axis) <= numpy.max(interval1, axis=axis),
            numpy.max(interval1, axis=axis) <= numpy.max(interval2, axis=axis),
        ),
        numpy.bitwise_and(
            numpy.min(interval2, axis=axis) <= numpy.min(interval1, axis=axis),
            numpy.min(interval1, axis=axis) <= numpy.max(interval2, axis=axis),
        ),
    )

    if oneway:
        return first_check
    else:
        return first_check | check_interval_overlap(interval2, interval1, oneway=True, axis=axis)


def winsorize_dataframe(df, **limits):
    """Winsorizes columns in a dataframe

    Parameters
    ----------
    dataframe : pandas.DataFrame
        The data to be modified (a copy is created).
    **limits : optional kwargs (floats, or two-tuples of floats)
        Optional key-value pairs of column names and (two-tuples of)
        floats that are the windsor limits to be applied to the
        respective column. Values should be between 0 and 1.

    Returns
    -------
    winsored_df : pandas.DataFrame
        The modified dataframe.

    See also
    --------
    scipy.stats.mstats.winsorize

    """
    newcols = {
        colname: stats.mstats.winsorize(df[colname], limits=limit)
        for colname, limit in limits.items()
    }

    return df.assign(**newcols)


def remove_outliers(x, factor=1.5):
    """Removes outliers from an array based on a scaling of the
    interquartile range (IQR).

    Parameters
    ----------
    x : array-like
    factor : float
        The scaling factor for the IQR

    Returns
    -------
    cleaned : numpy.array
        Copy of *x* with the outliers removed

    Example
    -------
    >>> import numpy
    >>> from wqio import utils
    >>> numpy.random.seed(0)
    >>> x = numpy.random.normal(0, 4, size=37) # shape is (37,)
    >>> utils.remove_outliers(x).shape
    (35,)

    """
    pctl25 = numpy.percentile(x, 25)
    pctl75 = numpy.percentile(x, 75)
    IQR = pctl75 - pctl25
    return x[(x >= pctl25 - (factor * IQR)) & (x <= pctl75 + (factor * IQR))]


def _comp_stat_generator(
    df, groupcols, pivotcol, rescol, statfxn, statname=None, pbarfxn=None, **statopts
):
    """Generator of records containing results of comparitive
    statistical functions.

    Parameters
    ----------
    df : pandas.DataFrame
    groupcols : list of string
        The columns on which the data will be groups. This should *not*
        include the column that defines the separate datasets that will
        be comparied.
    pivotcol : string
        The column that distinquishes the individual samples that are
        being compared.
    rescol : string
        The column that containes the values to compare.
    statfxn : callable
        Any function that takes to array-like datasets and returns a
        ``namedtuple`` with elements "statistic" and "pvalue". If the
        "pvalue" portion is not relevant, simply wrap the function such
        that it returns None or similar for it.
    statname : string, optional
        The name of the primary statistic being returned.
    **statopts : kwargs
        Additional options to be fed to ``statfxn``.

    Returns
    -------
    comp_stats : pandas.DataFrame

    """
    if not pbarfxn:
        pbarfxn = misc.no_op

    groupcols = validate.at_least_empty_list(groupcols)
    if statname is None:
        statname = "stat"

    groups = df.groupby(by=groupcols)
    for name, g in pbarfxn(groups):
        stations = g[pivotcol].unique()
        name = validate.at_least_empty_list(name)
        for _x, _y in pbarfxn(itertools.permutations(stations, 2)):
            row = dict(zip(groupcols, name))

            station_columns = [pivotcol + "_1", pivotcol + "_2"]
            row.update(dict(zip(station_columns, [_x, _y])))

            x = g.loc[g[pivotcol] == _x, rescol].values
            y = g.loc[g[pivotcol] == _y, rescol].values

            stat = statfxn(x, y, **statopts)
            row.update({statname: stat[0], "pvalue": stat.pvalue})
            yield row


def _paired_stat_generator(
    df, groupcols, pivotcol, rescol, statfxn, statname=None, pbarfxn=None, **statopts
):
    """Generator of records containing results of comparitive
    statistical functions specifically for paired data.

    Parameters
    ----------
    df : pandas.DataFrame
    groupcols : list of string
        The columns on which the data will be groups. This should *not*
        include the column that defines the separate datasets that will
        be comparied.
    pivotcol : string
        The column that distinquishes the individual samples that are
        being compared.
    rescol : string
        The column that containes the values to compare.
    statfxn : callable
        Any function that takes to array-like datasets and returns a
        ``namedtuple`` with elements "statistic" and "pvalue". If the
        "pvalue" portion is not relevant, simply wrap the function such
        that it returns None or similar for it.
    statname : string, optional
        The name of the primary statistic being returned.
    **statopts : kwargs
        Additional options to be fed to ``statfxn``.

    Returns
    -------
    comp_stats : pandsa.DataFrame

    """
    if not pbarfxn:
        pbarfxn = misc.no_op

    groupcols = validate.at_least_empty_list(groupcols)
    if statname is None:
        statname = "stat"

    groups = df.groupby(level=groupcols)[rescol]
    for name, g in pbarfxn(groups):
        stations = g.columns.tolist()
        name = validate.at_least_empty_list(name)
        for _x, _y in pbarfxn(itertools.permutations(stations, 2)):
            row = dict(zip(groupcols, name))

            station_columns = [pivotcol + "_1", pivotcol + "_2"]
            row.update(dict(zip(station_columns, [_x, _y])))

            _df = g[[_x, _y]].dropna()

            x = _df[_x].values
            y = _df[_y].values

            stat = statfxn(x, y, **statopts)
            row.update({statname: stat[0], "pvalue": stat.pvalue})
            yield row



import numpy
from probscale.algo import _estimate_from_fit

from wqio import utils


class TheilSenFit:
    def __init__(self, infl, effl, log_infl=False, log_effl=False, **theil_opts):
        """Theil-Sen Fit object

        Parameters
        ----------
        infl, effl : array-like
            Influent and effluent result to regress.
        log_infl, log_effl : bool
            When True, performs the fit on log-transformed data.

        See also
        --------
        wqio.utils.compute_theilslope

        """
        self.influent_data = infl
        self.effluent_data = effl

        self.log_infl = log_infl
        self.log_effl = log_effl

        if self.log_infl:
            self._infl_trans_in = numpy.log
            self._infl_trans_out = numpy.exp
        else:
            self._infl_trans_in = utils.no_op
            self._infl_trans_out = utils.no_op

        if self.log_effl:
            self._effl_trans_in = numpy.log
            self._effl_trans_out = numpy.exp
        else:
            self._effl_trans_in = utils.no_op
            self._effl_trans_out = utils.no_op

        self.theil_opts = theil_opts
        self._theil_stats = None

    @property
    def infl(self):
        return self._infl_trans_in(self.influent_data)

    @property
    def effl(self):
        return self._effl_trans_in(self.effluent_data)

    @property
    def theil_stats(self):
        if self._theil_stats is None:
            self._theil_stats = utils.compute_theilslope(
                y=self.effl, x=self.infl, **self.theil_opts
            )

        return self._theil_stats

    @property
    def med_slope(self):
        return self.theil_stats[0]

    @property
    def intercept(self):
        return self.theil_stats[1]

    @property
    def low_slope(self):
        return self.theil_stats[2]

    @property
    def high_slope(self):
        return self.theil_stats[3]

    @property
    def x_fit(self):
        xmin = self.influent_data.min()
        xmax = self.effluent_data.max()
        if self.log_infl:
            return numpy.logspace(numpy.log10(xmin), numpy.log10(xmax))
        else:
            return numpy.linspace(xmin, xmax)

    @property
    def med_estimate(self):
        return _estimate_from_fit(
            self.influent_data,
            self.med_slope,
            self.intercept,
            xlog=self.log_infl,
            ylog=self.log_effl,
        )

    @property
    def errors(self):
        return self._effl_trans_in(self.effluent_data) - self._effl_trans_in(self.med_estimate)

    @property
    def MAD(self):
        return numpy.median(self._effl_trans_out(numpy.abs(self.errors)))

    @property
    def BCF(self):
        return numpy.mean(self._effl_trans_out(utils.remove_outliers(self.errors)))


def all_theil(infl, effl, **theil_opts):
    """
    Convenience function to create the relevant TheilSenFits objects
    for a dataset in various log/linear spaces. The case with linear
    influent data, but logarithmic effluent data is omitted.

    Parameters
    -----------
    infl, effl : array-like
        Influent and effluent result to regress.

    Returns
    -------
    all_tr : list of TheilSenFit instances
    best_tr : TheilSenFit
        The TheilSenFit with the smallest mean absolute deviatation
        in *all_tr*.

    """

    all_tsf = [
        TheilSenFit(infl, effl, log_infl=False, log_effl=False, **theil_opts),
        TheilSenFit(infl, effl, log_infl=True, log_effl=False, **theil_opts),
        TheilSenFit(infl, effl, log_infl=True, log_effl=True, **theil_opts),
    ]
    best_tsf = min(all_tsf, key=lambda tr: tr.MAD)
    return all_tsf, best_tsf



import numpy
import numpy.testing as nptest
import pytest

from wqio import bootstrap
from wqio.tests import helpers


@pytest.fixture
def testdata():
    return helpers.getTestROSData()["res"].values


def test__acceleration(testdata):
    known_acceleration = -0.024051865664929263
    assert abs(bootstrap._acceleration(testdata) - known_acceleration) < 0.00001


def test__make_boot_index():
    result = bootstrap._make_boot_index(5, 5000)
    assert result.shape == (5000, 5)
    assert result.min() == 0
    assert result.max() == 4


@helpers.seed
@pytest.mark.parametrize(
    ("bootstrapper", "known_ci"),
    [
        (bootstrap.BCA, numpy.array([8.686, 11.661])),
        (bootstrap.percentile, numpy.array([8.670, 11.647])),
    ],
)
def test_bootstrappers(testdata, bootstrapper, known_ci):
    ci = bootstrapper(testdata, numpy.mean, 10000, 0.10)
    nptest.assert_array_almost_equal(known_ci, ci, decimal=3)


@helpers.seed
@pytest.mark.parametrize("uselog", [True, False])
def test_fit(uselog):
    N = 10
    x = numpy.arange(1, N + 1, dtype=float)
    y = numpy.array([4.527, 3.519, 9.653, 8.036, 10.805, 14.329, 13.508, 11.822, 13.281, 10.410])
    bsfit = bootstrap.fit(
        x, y, numpy.polyfit, niter=2000, xlog=uselog, ylog=uselog, deg=1, full=False
    )

    nptest.assert_array_almost_equal(bsfit.xhat, x[numpy.argsort(x)])

    expected = {
        False: {
            "yhat": numpy.array(
                [
                    5.841,
                    6.763,
                    7.685,
                    8.606,
                    9.528,
                    10.450,
                    11.371,
                    12.293,
                    13.215,
                    14.136,
                ]
            ),
            "lower": numpy.array(
                [
                    3.449,
                    4.855,
                    6.166,
                    7.181,
                    8.088,
                    8.954,
                    9.816,
                    10.561,
                    11.119,
                    11.591,
                ]
            ),
            "upper": numpy.array(
                [
                    9.948,
                    10.233,
                    10.562,
                    10.913,
                    11.542,
                    12.527,
                    13.901,
                    15.545,
                    17.218,
                    18.940,
                ]
            ),
        },
        True: {
            "yhat": numpy.array(
                [
                    4.004,
                    5.858,
                    7.318,
                    8.570,
                    9.686,
                    10.706,
                    11.651,
                    12.537,
                    13.374,
                    14.170,
                ]
            ),
            "lower": numpy.array(
                [
                    2.172,
                    3.961,
                    5.636,
                    7.079,
                    8.255,
                    9.218,
                    10.060,
                    10.760,
                    11.280,
                    11.660,
                ]
            ),
            "upper": numpy.array(
                [
                    8.048,
                    9.080,
                    9.882,
                    10.589,
                    11.429,
                    12.494,
                    13.871,
                    15.438,
                    17.124,
                    18.713,
                ]
            ),
        },
    }
    nptest.assert_array_almost_equal(bsfit.yhat, expected[uselog]["yhat"], decimal=3)
    nptest.assert_array_almost_equal(bsfit.lower, expected[uselog]["lower"], decimal=3)
    nptest.assert_array_almost_equal(bsfit.upper, expected[uselog]["upper"], decimal=3)
    assert bsfit.xlog == uselog
    assert bsfit.ylog == uselog



import os

import numpy
import pandas
from matplotlib import pyplot


def dataset(fname):
    valid_names = ["bmpdata", "cvc", "nsqd"]
    fname, ext = os.path.splitext(fname)
    if fname.lower() not in valid_names:
        msg = "filename '{}' not one of {}"
        raise ValueError(msg.format(fname, valid_names))

    return fname.lower() + ".zip"


def timestamp(datelike):
    """Converts datetime-like objects to pandas.Timestamp.
    Pretty miuch a direct pass through, but give a slighly
    more informative error message.

    Parameters
    ----------
    datelike : datetime.datetime, string, or Timestamp
        The value to be coerced into the pandas.Timestamp.

    Returns
    -------
    tstamp : pandas.Timestamp
        Coerced value.

    """

    try:
        tstamp = pandas.Timestamp(datelike)
    except ValueError:
        msg = f"{datelike} could not be coerced into a pandas.Timestamp"
        raise ValueError(msg)

    return tstamp


def axes(ax, fallback="new"):
    """Checks if a value if an Axes. If None, a new one is created or
    the 'current' one is found.

    Parameters
    ----------
    ax : matplotlib.Axes or None
        The value to be tested.
    fallback : str, optional
        If ax is None. ``fallback='new'`` will create a new axes and
        figure. The other option is ``fallback='current'``, where the
        "current" axes are return via ``pyplot.gca()``

    Returns
    -------
    fig : matplotlib.Figure
    ax : matplotlib.Axes

    """

    if ax is None:
        if fallback == "new":
            fig, ax = pyplot.subplots()
        elif fallback == "current":
            ax = pyplot.gca()
            fig = ax.figure
        else:
            raise ValueError("fallback must be either 'new' or 'current'")

    elif isinstance(ax, pyplot.Axes):
        fig = ax.figure

    else:
        msg = "`ax` must be a matplotlib Axes instance or None"
        raise ValueError(msg)

    return fig, ax


def single_value_in_index(df, index_level):
    """Confirms that a given level of a dataframe's index only has
    one unique value. Useful for confirming consistent units. Raises
    error if level is not a single value. Returns unique value of the
    index level.

    Parameters
    ----------
    df : pandas.DataFrame
        Dataframe whose index will be inspected.
    index_level : int or string
        Level of the dataframe's index that should only have one unique
        value.

    Returns
    -------
    uniqueval
        The unique value of the index.

    """

    index = numpy.unique(df.index.get_level_values(index_level).tolist())
    if index.shape != (1,):
        raise ValueError(f'index level "{index_level}" is not unique.')

    return index[0]


def at_least_empty_list(value):
    if isinstance(value, numpy.ndarray):
        value = value.tolist()
    elif numpy.isscalar(value) and value != "":
        value = [value]
    elif not value:
        value = []

    return value


def at_least_empty_dict(value, **kwargs):
    if value is None or value == "":
        value = {}
    elif not isinstance(value, dict):
        raise ValueError(f"{value} cannot be a dictionary")
    else:
        value = value.copy()

    value.update(**kwargs)

    return value


def fit_arguments(arg, argname):
    valid_args = ["x", "y", "both", None]
    if arg not in valid_args:
        msg = "Valid value for {} ({}). Must be on of {}"
        raise ValueError(msg.format(argname, arg, valid_args))

    return arg



import warnings

from pkg_resources import resource_filename

from wqio.tests.helpers import requires

try:
    import pytest
except ImportError:
    pytest = None


@requires(pytest, "pytest")
def test(*args):
    options = [resource_filename("wqio", "")]
    options.extend(list(args))
    return pytest.main(options)


@requires(pytest, "pytest")
def teststrict(*args):
    options = ["--mpl", "--doctest-modules", *list(args)]
    return test(*list(set(options)))


@requires(pytest, "pytest")
def test_nowarnings(*args):
    with warnings.catch_warnings():
        warnings.simplefilter("error")
        return teststrict(*args)



import warnings

import numpy
import pandas
import seaborn
from matplotlib import dates, gridspec, pyplot
from pandas.plotting import register_matplotlib_converters

from wqio import utils, validate, viz

register_matplotlib_converters()

SEC_PER_MINUTE = 60.0
MIN_PER_HOUR = 60.0
HOUR_PER_DAY = 24.0
SEC_PER_HOUR = SEC_PER_MINUTE * MIN_PER_HOUR
SEC_PER_DAY = SEC_PER_HOUR * HOUR_PER_DAY


def _wet_first_row(df, wetcol, diffcol):
    # make sure that if the first record is associated with the first
    # storm if it's wet
    firstrow = df.iloc[0]
    if firstrow[wetcol]:
        df.loc[firstrow.name, diffcol] = 1

    return df


def _wet_window_diff(is_wet, ie_periods):
    return (
        is_wet.rolling(int(ie_periods), min_periods=1)
        .apply(lambda window: window.any(), raw=False)
        .diff()
    )


def parse_storm_events(
    data,
    intereventHours,
    outputfreqMinutes,
    precipcol=None,
    inflowcol=None,
    outflowcol=None,
    baseflowcol=None,
    stormcol="storm",
    debug=False,
):
    """Parses the hydrologic data into distinct storms.

    In this context, a storm is defined as starting whenever the
    hydrologic records shows non-zero precipitation or [in|out]flow
    from the BMP after a minimum inter-event dry period duration
    specified in the the function call. The storms ends the observation
    *after* the last non-zero precipitation or flow value.

    Parameters
    ----------
    data : pandas.DataFrame
    intereventHours : float
        The Inter-Event dry duration (in hours) that classifies the
        next hydrlogic activity as a new event.
    precipcol : string, optional (default = None)
        Name of column in `hydrodata` containing precipiation data.
    inflowcol : string, optional (default = None)
        Name of column in `hydrodata` containing influent flow data.
    outflowcol : string, optional (default = None)
        Name of column in `hydrodata` containing effluent flow data.
    baseflowcol : string, optional (default = None)
        Name of column in `hydrodata` containing boolean indicating
        which records are considered baseflow.
    stormcol : string (default = 'storm')
        Name of column in `hydrodata` indentifying distinct storms.
    debug : bool (default = False)
        If True, diagnostic columns will not be dropped prior to
        returning the dataframe of parsed_storms.

    Writes
    ------
    None

    Returns
    -------
    parsed_storms : pandas.DataFrame
        Copy of the origin `hydrodata` DataFrame, but resampled to a
        fixed frequency, columns possibly renamed, and a `storm` column
        added to denote the storm to which each record belongs. Records
        where `storm` == 0 are not a part of any storm.

    """

    # pull out the rain and flow data
    if precipcol is None:
        precipcol = "precip"
        data.loc[:, precipcol] = numpy.nan

    if inflowcol is None:
        inflowcol = "inflow"
        data.loc[:, inflowcol] = numpy.nan

    if outflowcol is None:
        outflowcol = "outflow"
        data.loc[:, outflowcol] = numpy.nan

    if baseflowcol is None:
        baseflowcol = "baseflow"
        data.loc[:, baseflowcol] = False

    # bool column where True means there's rain or flow of some kind
    water_columns = [inflowcol, outflowcol, precipcol]
    cols_to_use = water_columns + [baseflowcol]

    agg_dict = {
        precipcol: numpy.sum,
        inflowcol: numpy.mean,
        outflowcol: numpy.mean,
        baseflowcol: numpy.any,
    }

    freq = pandas.offsets.Minute(outputfreqMinutes)
    ie_periods = int(MIN_PER_HOUR / freq.n * intereventHours)

    # periods between storms are where the cumulative number
    # of storms that have ended are equal to the cumulative
    # number of storms that have started.
    # Stack Overflow: http://tinyurl.com/lsjkr9x
    res = (
        data.resample(freq)
        .agg(agg_dict)
        .loc[:, lambda df: df.columns.isin(cols_to_use)]
        .assign(__wet=lambda df: numpy.any(df[water_columns] > 0, axis=1) & ~df[baseflowcol])
        .assign(__windiff=lambda df: _wet_window_diff(df["__wet"], ie_periods))
        .pipe(_wet_first_row, "__wet", "__windiff")
        .assign(__event_start=lambda df: df["__windiff"] == 1)
        .assign(__event_end=lambda df: df["__windiff"].shift(-1 * ie_periods) == -1)
        .assign(__storm=lambda df: df["__event_start"].cumsum())
        .assign(
            storm=lambda df: numpy.where(
                df["__storm"] == df["__event_end"].shift(2).cumsum(),
                0,  # inter-event periods marked as zero
                df["__storm"],  # actual events keep their number
            )
        )
    )

    if not debug:
        res = res.loc[:, res.columns.map(lambda c: not c.startswith("__"))]

    return res


class Storm:
    """Object representing a storm event

    Parameters
    ----------
    dataframe : pandas.DataFrame
        A datetime-indexed Dataframe containing all of the hydrologic
        data and am interger column indentifying distinct storms.
    stormnumber : int
        The storm we care about.
    precipcol, inflowcol, outflow, tempcol, stormcol : string, optional
        Names for columns representing each hydrologic quantity.
    freqMinutes : float (default = 5)
        The time period, in minutes, between observations.
    volume_conversion : float, optional (default = 1)
        Conversion factor to go from flow to volume for a single
        observation.

    """

    # TODO: rename freqMinutes to periodMinutes
    def __init__(
        self,
        dataframe,
        stormnumber,
        precipcol="precip",
        inflowcol="inflow",
        outflowcol="outflow",
        tempcol="temp",
        stormcol="storm",
        freqMinutes=5,
        volume_conversion=1,
    ):
        self.inflowcol = inflowcol
        self.outflowcol = outflowcol
        self.precipcol = precipcol
        self.tempcol = tempcol
        self.stormnumber = stormnumber
        self.freqMinutes = freqMinutes
        self.volume_conversion = volume_conversion * SEC_PER_MINUTE * self.freqMinutes

        # basic data
        self.data = dataframe[dataframe[stormcol] == self.stormnumber].copy()
        self.hydrofreq_label = f"{self.freqMinutes} min"

        # tease out start/stop info
        self.start = self.data.index[0]
        self.end = self.data.index[-1]
        self._season = utils.get_season(self.start)

        # storm duration (hours)
        duration = self.end - self.start
        self.duration_hours = duration.total_seconds() / SEC_PER_HOUR

        # antecedent dry period (hours)
        if self.stormnumber > 1:
            prev_storm_mask = dataframe[stormcol] == self.stormnumber - 1
            previous_end = dataframe[prev_storm_mask].index[-1]
            antecedent_timedelta = self.start - previous_end
            self.antecedent_period_days = antecedent_timedelta.total_seconds() / SEC_PER_DAY
        else:
            self.antecedent_period_days = numpy.nan

        # quantities
        self._precip = None
        self._inflow = None
        self._outflow = None

        # starts and stop
        self._precip_start = None
        self._precip_end = None
        self._inflow_start = None
        self._inflow_end = None
        self._outflow_start = None
        self._outflow_end = None

        # peaks
        self._peak_precip_intensity = None
        self._peak_inflow = None
        self._peak_outflow = None

        # times of peaks
        self._peak_precip_intensity_time = None
        self._peak_inflow_time = None
        self._peak_outflow_time = None
        self._peak_lag_hours = None

        # centroids
        self._centroid_precip_time = None
        self._centroid_inflow_time = None
        self._centroid_outflow_time = None
        self._centroid_lag_hours = None

        # totals
        self._total_precip_depth = None
        self._total_inflow_volume = None
        self._total_outflow_volume = None

        self.meta = {
            self.outflowcol: {
                "name": "Flow (calculated, L/s)",
                "ylabel": "Effluent flow (L/s)",
                "color": "CornFlowerBlue",
                "linewidth": 1.5,
                "alpha": 0.5,
                "ymin": 0,
            },
            self.inflowcol: {
                "name": "Inflow (estimated, L/s)",
                "ylabel": "Estimated influent flow (L/s)",
                "color": "Maroon",
                "linewidth": 1.5,
                "alpha": 0.5,
                "ymin": 0,
            },
            self.precipcol: {
                "name": "Precip (mm)",
                "ylabel": "%s Precip.\nDepth (mm)" % self.hydrofreq_label,
                "color": "DarkGreen",
                "linewidth": 1.5,
                "alpha": 0.4,
                "ymin": 0,
            },
            self.tempcol: {
                "name": "Air Temp (deg C)",
                "ylabel": "Air Temperature (deg. C)",
                "color": "DarkGoldenRod",
                "linewidth": 1.5,
                "alpha": 0.5,
                "ymin": None,
            },
        }

        self._summary_dict = None

    @property
    def precip(self):
        if self._precip is None:
            if self.precipcol is not None:
                self._precip = self.data[self.data[self.precipcol] > 0][self.precipcol]
            else:
                self._precip = numpy.array([])
        return self._precip

    @property
    def inflow(self):
        if self._inflow is None:
            if self.inflowcol is not None:
                self._inflow = self.data[self.data[self.inflowcol] > 0][self.inflowcol]
            else:
                self._inflow = numpy.array([])
        return self._inflow

    @property
    def outflow(self):
        if self._outflow is None:
            if self.outflowcol is not None:
                self._outflow = self.data[self.data[self.outflowcol] > 0][self.outflowcol]
            else:
                self._outflow = numpy.array([])
        return self._outflow

    @property
    def has_precip(self):
        return self.precip.shape[0] > 0

    @property
    def has_inflow(self):
        return self.inflow.shape[0] > 0

    @property
    def has_outflow(self):
        return self.outflow.shape[0] > 0

    @property
    def season(self):
        return self._season

    @season.setter
    def season(self, value):
        self._season = value

    # starts and stops
    @property
    def precip_start(self):
        if self._precip_start is None and self.has_precip:
            self._precip_start = self._get_event_time(self.precipcol, "start")
        return self._precip_start

    @property
    def precip_end(self):
        if self._precip_end is None and self.has_precip:
            self._precip_end = self._get_event_time(self.precipcol, "end")
        return self._precip_end

    @property
    def inflow_start(self):
        if self._inflow_start is None and self.has_inflow:
            self._inflow_start = self._get_event_time(self.inflowcol, "start")
        return self._inflow_start

    @property
    def inflow_end(self):
        if self._inflow_end is None and self.has_inflow:
            self._inflow_end = self._get_event_time(self.inflowcol, "end")
        return self._inflow_end

    @property
    def outflow_start(self):
        if self._outflow_start is None and self.has_outflow:
            self._outflow_start = self._get_event_time(self.outflowcol, "start")
        return self._outflow_start

    @property
    def outflow_end(self):
        if self._outflow_end is None and self.has_outflow:
            self._outflow_end = self._get_event_time(self.outflowcol, "end")
        return self._outflow_end

    @property
    def _peak_depth(self):
        if self.has_precip:
            return self.precip.max()

    @property
    def peak_precip_intensity(self):
        if self._peak_precip_intensity is None and self.has_precip:
            self._peak_precip_intensity = self._peak_depth * MIN_PER_HOUR / self.freqMinutes
        return self._peak_precip_intensity

    @property
    def peak_inflow(self):
        if self._peak_inflow is None and self.has_inflow:
            self._peak_inflow = self.inflow.max()
        return self._peak_inflow

    @property
    def peak_outflow(self):
        if self._peak_outflow is None and self.has_outflow:
            self._peak_outflow = self.outflow.max()
        return self._peak_outflow

    @property
    def total_precip_depth(self):
        if self._total_precip_depth is None and self.has_precip:
            self._total_precip_depth = self.data[self.precipcol].sum()
        return self._total_precip_depth

    @property
    def total_inflow_volume(self):
        if self._total_inflow_volume is None and self.has_inflow:
            self._total_inflow_volume = self.data[self.inflowcol].sum() * self.volume_conversion
        return self._total_inflow_volume

    @property
    def total_outflow_volume(self):
        if self._total_outflow_volume is None and self.has_outflow:
            self._total_outflow_volume = self.data[self.outflowcol].sum() * self.volume_conversion
        return self._total_outflow_volume

    @property
    def centroid_precip_time(self):
        if self._centroid_precip_time is None and self.has_precip:
            self._centroid_precip_time = self._compute_centroid(self.precipcol)
        return self._centroid_precip_time

    @property
    def centroid_inflow_time(self):
        if self._centroid_inflow_time is None and self.has_inflow:
            self._centroid_inflow_time = self._compute_centroid(self.inflowcol)
        return self._centroid_inflow_time

    @property
    def centroid_outflow_time(self):
        if self._centroid_outflow_time is None and self.has_outflow:
            self._centroid_outflow_time = self._compute_centroid(self.outflowcol)
        return self._centroid_outflow_time

    @property
    def centroid_lag_hours(self):
        if (
            self._centroid_lag_hours is None
            and self.centroid_outflow_time is not None
            and self.centroid_inflow_time is not None
        ):
            self._centroid_lag_hours = (
                self.centroid_outflow_time - self.centroid_inflow_time
            ).total_seconds() / SEC_PER_HOUR
        return self._centroid_lag_hours

    @property
    def peak_precip_intensity_time(self):
        if self._peak_precip_intensity_time is None and self.has_precip:
            PI_selector = self.data[self.precipcol] == self._peak_depth
            self._peak_precip_intensity_time = self.data[PI_selector].index[0]
        return self._peak_precip_intensity_time

    @property
    def peak_inflow_time(self):
        if self._peak_inflow_time is None and self.has_inflow:
            PInf_selector = self.data[self.inflowcol] == self.peak_inflow
            self._peak_inflow_time = self.data[PInf_selector].index[0]
        return self._peak_inflow_time

    @property
    def peak_outflow_time(self):
        if self._peak_outflow_time is None and self.has_outflow:
            PEff_selector = self.data[self.outflowcol] == self.peak_outflow
            if PEff_selector.sum() > 0:
                self._peak_outflow_time = self.data[PEff_selector].index[0]
        return self._peak_outflow_time

    @property
    def peak_lag_hours(self):
        if (
            self._peak_lag_hours is None
            and self.peak_outflow_time is not None
            and self.peak_inflow_time is not None
        ):
            time_delta = self.peak_outflow_time - self.peak_inflow_time
            self._peak_lag_hours = time_delta.total_seconds() / SEC_PER_HOUR
        return self._peak_lag_hours

    @property
    def summary_dict(self):
        if self._summary_dict is None:
            self._summary_dict = {
                "Storm Number": self.stormnumber,
                "Antecedent Days": self.antecedent_period_days,
                "Start Date": self.start,
                "End Date": self.end,
                "Duration Hours": self.duration_hours,
                "Peak Precip Intensity": self.peak_precip_intensity,
                "Total Precip Depth": self.total_precip_depth,
                "Total Inflow Volume": self.total_inflow_volume,
                "Peak Inflow": self.peak_inflow,
                "Total Outflow Volume": self.total_outflow_volume,
                "Peak Outflow": self.peak_outflow,
                "Peak Lag Hours": self.peak_lag_hours,
                "Centroid Lag Hours": self.centroid_lag_hours,
                "Season": self.season,
            }

        return self._summary_dict

    def is_small(self, minprecip=0.0, mininflow=0.0, minoutflow=0.0):
        """Determines whether a storm can be considered "small".

        Parameters
        ----------
        minprecip, mininflow, minoutflow : float, optional (default = 0)
            The minimum amount of each hydrologic quantity below which a
            storm can be considered "small".

        Returns
        -------
        storm_is_small : bool
            True if the storm is considered small.

        """

        storm_is_small = (
            (self.total_precip_depth is not None and self.total_precip_depth < minprecip)
            or (self.total_inflow_volume is not None and self.total_inflow_volume < mininflow)
            or (self.total_outflow_volume is not None and self.total_outflow_volume < minoutflow)
        )
        return storm_is_small

    def _get_event_time(self, column, bound):
        index_map = {"start": 0, "end": -1}
        quantity = self.data[self.data[column] > 0]
        if quantity.shape[0] == 0:
            warnings.warn(f"Storm has no {column}", UserWarning)
        else:
            return quantity.index[index_map[bound]]

    def _get_max_quantity(self, column):
        return self.data[column].max()

    def _compute_centroid(self, column):
        # ordinal time index of storm
        time_idx = [dates.date2num(idx.to_pydatetime()) for idx in self.data.index.tolist()]

        centroid = numpy.sum(self.data[column] * time_idx) / numpy.sum(self.data[column])

        if numpy.isnan(centroid):
            return None
        else:
            return pandas.Timestamp(dates.num2date(centroid)).tz_convert(None)

    def _plot_centroids(self, ax, yfactor=0.5):
        artists = []
        labels = []
        y_val = yfactor * ax.get_ylim()[1]

        if self.centroid_precip is not None:
            ax.plot(
                [self.centroid_precip],
                [y_val],
                color="DarkGreen",
                marker="o",
                linestyle="none",
                zorder=20,
                markersize=6,
            )
            artists.append(
                pyplot.Line2D(
                    [0],
                    [0],
                    marker=".",
                    markersize=6,
                    linestyle="none",
                    color="DarkGreen",
                )
            )
            labels.append("Precip. centroid")

        if self.centroid_flow is not None:
            ax.plot(
                [self.centroid_flow],
                [y_val],
                color="CornflowerBlue",
                marker="s",
                linestyle="none",
                zorder=20,
                markersize=6,
            )
            artists.append(
                pyplot.Line2D(
                    [0],
                    [0],
                    marker="s",
                    markersize=6,
                    linestyle="none",
                    color="CornflowerBlue",
                )
            )
            labels.append("Effluent centroid")

        if self.centroid_precip is not None and self.centroid_flow is not None:
            ax.annotate(
                "",
                (self.centroid_flow, y_val),
                arrowprops=dict(arrowstyle="-|>"),
                xytext=(self.centroid_precip, y_val),
            )

        return artists, labels

    def plot_hydroquantity(self, quantity, ax=None, label=None, otherlabels=None, artists=None):
        """Draws a hydrologic quantity to a matplotlib axes.

        Parameters
        ----------
        quantity : string
            Column name of the quantity you want to plot.
        ax : matplotlib axes object, optional
            The axes on which the data will be plotted. If None, a new
            one will be created.
        label : string, optional
            How the series should be labeled in the figure legend.
        otherlabels : list of strings, optional
            A list of other legend labels that have already been plotted
            to ``ax``. If provided, ``label`` will be appended. If not
            provided, and new list will be created.
        artists : list of matplotlib artists, optional
            A list of other legend items that have already been plotted
            to ``ax``. If provided, the artist created will be appended.
            If not provided, and new list will be created.

        Returns
        -------
        fig : matplotlib.Figure
            The figure containing the plot.
        labels : list of strings
            Labels to be included in a legend for the figure.
        artists : list of matplotlib artists
            Symbology for the figure legend.

        """

        # setup the figure
        fig, ax = validate.axes(ax)

        if label is None:
            label = quantity

        # select the plot props based on the column
        try:
            meta = self.meta[quantity]
        except KeyError:
            raise KeyError(f"{quantity} not available")

        # plot the data
        self.data[quantity].fillna(0).plot(
            ax=ax, kind="area", color=meta["color"], alpha=meta["alpha"], zorder=5
        )

        if artists is not None:
            proxy = pyplot.Rectangle(
                (0, 0), 1, 1, facecolor=meta["color"], linewidth=0, alpha=meta["alpha"]
            )
            artists.append(proxy)
        if otherlabels is not None:
            otherlabels.append(label)

        return fig, otherlabels, artists

    def summaryPlot(
        self,
        axratio=2,
        filename=None,
        showLegend=True,
        precip=True,
        inflow=True,
        outflow=True,
        figopts={},
        serieslabels={},
    ):
        """
        Creates a figure showing the hydrlogic record (flow and
            precipitation) of the storm

        Input:
            axratio : optional float or int (default = 2)
                Relative height of the flow axis compared to the
                precipiation axis.

            filename : optional string (default = None)
                Filename to which the figure will be saved.

            **figwargs will be passed on to `pyplot.Figure`

        Writes:
            Figure of flow and precipitation for a storm

        Returns:
            None
        """
        fig = pyplot.figure(**figopts)
        gs = gridspec.GridSpec(nrows=2, ncols=1, height_ratios=[1, axratio], hspace=0.12)
        rainax = fig.add_subplot(gs[0])
        rainax.yaxis.set_major_locator(pyplot.MaxNLocator(5))
        flowax = fig.add_subplot(gs[1], sharex=rainax)

        # create the legend proxy artists
        artists = []
        labels = []

        # in the label assignment: `serieslabels.pop(item, item)` might
        # seem odd. What it does is looks for a label (value) in the
        # dictionary with the key equal to `item`. If there is no valur
        # for that key in the dictionary the `item` itself is returned.
        # so if there's nothing called "test" in mydict,
        # `mydict.pop("test", "test")` returns `"test"`.
        if inflow:
            fig, labels, artists = self.plot_hydroquantity(
                self.inflowcol,
                ax=flowax,
                label=serieslabels.pop(self.inflowcol, self.inflowcol),
                otherlabels=labels,
                artists=artists,
            )

        if outflow:
            fig, labels, arti = self.plot_hydroquantity(
                self.outflowcol,
                ax=flowax,
                label=serieslabels.pop(self.outflowcol, self.outflowcol),
                otherlabels=labels,
                artists=artists,
            )

        if precip:
            fig, labels, arti = self.plot_hydroquantity(
                self.precipcol,
                ax=rainax,
                label=serieslabels.pop(self.precipcol, self.precipcol),
                otherlabels=labels,
                artists=artists,
            )
            rainax.invert_yaxis()

        if showLegend:
            leg = rainax.legend(
                artists,
                labels,
                fontsize=7,
                ncol=1,
                markerscale=0.75,
                frameon=False,
                loc="lower right",
            )
            leg.get_frame().set_zorder(25)
            _leg = [leg]
        else:
            _leg = None

        seaborn.despine(ax=rainax, bottom=True, top=False)
        seaborn.despine(ax=flowax)
        flowax.set_xlabel("")
        rainax.set_xlabel("")

        if filename is not None:
            fig.savefig(
                filename,
                dpi=300,
                transparent=True,
                bbox_inches="tight",
                bbox_extra_artists=_leg,
            )

        return fig, artists, labels


class HydroRecord:
    """Class representing an entire hydrologic record.

    Parameters
    ----------
    hydrodata : pandas.DataFrame
        DataFrame of hydrologic data of the storm. Should contain
        a unique index of type pandas.DatetimeIndex.
    precipcol : string, optional (default = None)
        Name of column in `hydrodata` containing precipiation data.
    inflowcol : string, optional (default = None)
        Name of column in `hydrodata` containing influent flow data.
    outflowcol : string, optional (default = None)
        Name of column in `hydrodata` containing effluent flow data.
    baseflowcol : string, optional (default = None)
        Name of column in `hydrodata` containing boolean indicating
        which records are considered baseflow.
    stormcol : string (default = 'storm')
        Name of column in `hydrodata` indentifying distinct storms.
    minprecip, mininflow, minoutflow : float, optional (default = 0)
        The minimum amount of each hydrologic quantity below which a
        storm can be considered "small".
    outputfreqMinutes : int, optional (default = 10)
        The default frequency (minutes) to which all data will be
        resampled. Precipitation data will be summed up across '
        multiple timesteps during resampling, while flow will be
        averaged.
    intereventHours : int, optional (default = 6)
        The dry duration (no flow or rain) required to signal the end of
        a storm.
    volume_conversion : float, optional (default = 1)
        Conversion factor to go from flow to volume for a single
        observation.
    stormclass : object, optional
        Defaults to wqio.hydro.Storm. Can be a subclass of that in cases
        where custom functionality is needed.
    lowmem : bool (default = False)
        If True, all dry observations are removed from the dataframe.

    """

    # TODO: rename `outputfreqMinutes` to `outputPeriodMinutes`
    def __init__(
        self,
        hydrodata,
        precipcol=None,
        inflowcol=None,
        outflowcol=None,
        baseflowcol=None,
        tempcol=None,
        stormcol="storm",
        minprecip=0.0,
        mininflow=0.0,
        minoutflow=0.0,
        outputfreqMinutes=10,
        intereventHours=6,
        volume_conversion=1,
        stormclass=None,
        lowmem=False,
    ):
        # validate input
        if precipcol is None and inflowcol is None and outflowcol is None:
            msg = "`hydrodata` must have at least a precip or in/outflow column"
            raise ValueError(msg)

        self.stormclass = stormclass or Storm

        # static input
        self._raw_data = hydrodata
        self.precipcol = precipcol
        self.inflowcol = inflowcol
        self.outflowcol = outflowcol
        self.baseflowcol = baseflowcol
        self.stormcol = stormcol
        self.tempcol = tempcol
        self.outputfreq = pandas.offsets.Minute(outputfreqMinutes)
        self.intereventHours = intereventHours
        self.intereventPeriods = MIN_PER_HOUR / self.outputfreq.n * self.intereventHours
        self.minprecip = minprecip
        self.mininflow = mininflow
        self.minoutflow = minoutflow
        self.volume_conversion = volume_conversion
        self.lowmem = lowmem

        # properties
        self._data = None
        self._all_storms = None
        self._storms = None
        self._storm_stats = None

    @property
    def data(self):
        if self._data is None:
            self._data = self._define_storms()
            if self.lowmem:
                self._data = self._data[self._data[self.stormcol] != 0]

        return self._data

    @property
    def all_storms(self):
        if self._all_storms is None:
            self._all_storms = {}
            for storm_number in self.data[self.stormcol].unique():
                if storm_number > 0:
                    this_storm = self.stormclass(
                        self.data,
                        storm_number,
                        precipcol=self.precipcol,
                        inflowcol=self.inflowcol,
                        outflowcol=self.outflowcol,
                        tempcol=self.tempcol,
                        stormcol=self.stormcol,
                        volume_conversion=self.volume_conversion,
                        freqMinutes=self.outputfreq.n,
                    )
                    self._all_storms[storm_number] = this_storm

        return self._all_storms

    @property
    def storms(self):
        if self._storms is None:
            self._storms = {}
            for snum, storm in self.all_storms.items():
                is_small = storm.is_small(
                    minprecip=self.minprecip,
                    mininflow=self.mininflow,
                    minoutflow=self.minoutflow,
                )

                if not is_small:
                    self._storms[snum] = storm

        return self._storms

    @property
    def storm_stats(self):
        col_order = [
            "Storm Number",
            "Antecedent Days",
            "Season",
            "Start Date",
            "End Date",
            "Duration Hours",
            "Peak Precip Intensity",
            "Total Precip Depth",
            "Total Inflow Volume",
            "Peak Inflow",
            "Total Outflow Volume",
            "Peak Outflow",
            "Peak Lag Hours",
            "Centroid Lag Hours",
        ]
        if self._storm_stats is None:
            storm_stats = pandas.DataFrame([self.storms[sn].summary_dict for sn in self.storms])

            self._storm_stats = storm_stats[col_order]

        return self._storm_stats.sort_values(by=["Storm Number"]).reset_index(drop=True)

    def _define_storms(self, debug=False):
        parsed = parse_storm_events(
            self._raw_data,
            self.intereventHours,
            self.outputfreq.n,
            precipcol=self.precipcol,
            inflowcol=self.inflowcol,
            outflowcol=self.outflowcol,
            baseflowcol=self.baseflowcol,
            stormcol="storm",
            debug=debug,
        )
        return parsed

    def getStormFromTimestamp(self, timestamp, lookback_hours=0, smallstorms=False):
        """Get the storm associdated with a give (sample) date

        Parameters
        ----------
        timestamp : pandas.Timestamp
            The date/time for which to search within the hydrologic
            record.
        lookback_hours : positive int or float, optional (default = 0)
            If no storm is actively occuring at the provided timestamp,
            we can optionally look backwards in the hydrologic record a
            fixed amount of time (specified in hours). Negative values
            are ignored.
        smallstorms : bool, optional (default = False)
            If True, small storms will be included in the search.

        Returns
        -------
        storm_number : int
        storm : wqio.Storm

        """

        # santize date input
        timestamp = validate.timestamp(timestamp)

        # check lookback hours
        if lookback_hours < 0:
            raise ValueError("`lookback_hours` must be greater than 0")

        # initial search for the storm
        storm_number = int(self.data.loc[:timestamp, self.stormcol].iloc[-1])

        # look backwards if we have too
        if (storm_number == 0 or pandas.isnull(storm_number)) and lookback_hours != 0:
            lookback_time = timestamp - pandas.offsets.Hour(lookback_hours)
            storms = self.data.loc[lookback_time:timestamp, [self.stormcol]]
            storms = storms[storms > 0].dropna()

            storm_number = None if storms.shape[0] == 0 else int(storms.iloc[-1])

        # return storm_number and storms
        if smallstorms:
            return storm_number, self.all_storms.get(storm_number, None)
        else:
            return storm_number, self.storms.get(storm_number, None)

    def histogram(self, valuecol, bins, **factoropts):
        """Plot a faceted, categorical histogram of storms.

        Parameters
        ----------
        valuecol : str, optional
            The name of the column that should be categorized and plotted.
        bins : array-like, optional
            The right-edges of the histogram bins.
        factoropts : keyword arguments, optional
            Options passed directly to seaborn.factorplot

        Returns
        -------
        fig : seaborn.FacetGrid

        See also
        --------
        viz.categorical_histogram
        seaborn.factorplot

        """

        fg = viz.categorical_histogram(self.storm_stats, valuecol, bins, **factoropts)
        fg.figure.tight_layout()
        return fg


class DrainageArea:
    def __init__(self, total_area=1.0, imp_area=1.0, bmp_area=0.0):
        """A simple object representing the drainage area of a BMP.

        Units are not enforced, so keep them consistent yourself. The
        calculations available assume that the area of the BMP and the
        "total" area are mutually exclusive. In other words,
        the watershed outlet is at the BMP inlet.

        Parameters
        ----------
        total_area : float, optional (default = 1.0)
            The total geometric area of the BMP's catchment
        imp_area : float, optional (default = 1.0)
            The impervious area of the BMP's catchment
        bmp_area : float, optional (default = 0.0)
            The geometric area of the BMP itself.

        """

        self.total_area = float(total_area)
        self.imp_area = float(imp_area)
        self.bmp_area = float(bmp_area)

    def simple_method(self, storm_depth, volume_conversion=1.0, annual_factor=1.0):
        """
        Estimate runoff volume via Bob Pitt's Simple Method.

        Parameters
        ----------
        storm_depth : float
            Depth of the storm.
        volume_conversion : float, optional (default = 1.0)
            Conversion factor to go from [area units] * [depth units] to
            the desired [volume units]. If [area] = m^2, [depth] = mm,
            and [volume] = L, then `volume_conversion` = 1.
        annual_factor : float, optional (default = 1.0)
            The Simple Method's annual correction factor to account for
            small storms that do not produce runoff.

        Returns
        -------
        runoff_volume : float
            The volume of water entering the BMP immediately downstream
            of the drainage area.

        """

        # volumetric run off coneffiecient
        Rv = 0.05 + (0.9 * (self.imp_area / self.total_area))

        # run per unit storm depth
        drainage_conversion = Rv * self.total_area * volume_conversion
        bmp_conversion = self.bmp_area * volume_conversion

        # total runoff based on actual storm depth
        runoff_volume = (drainage_conversion * annual_factor + bmp_conversion) * storm_depth
        return runoff_volume



import numpy.testing as nptest
import pandas
import pytest
import scipy
from packaging.version import Version

from wqio.features import Dataset, Location
from wqio.tests import helpers

OLD_SCIPY = Version(scipy.version.version) < Version("0.19")
TOLERANCE = 0.05


@pytest.fixture(params=[True, False])
def location(request):
    data = helpers.getTestROSData()
    return Location(
        data,
        station_type="inflow",
        bsiter=1500,
        rescol="res",
        qualcol="qual",
        useros=request.param,
    )


@pytest.mark.parametrize(
    "attr",
    [
        "name",
        "station_type",
        "analysis_space",
        "rescol",
        "qualcol",
        "plot_marker",
        "scatter_marker",
        "hasData",
        "all_positive",
        "include",
        "exclude",
        "NUnique",
        "bsiter",
        "N",
        "ND",
    ],
)
def test_locations_strings_ints(location, attr):
    expected = {
        "name": "Influent",
        "station_type": "inflow",
        "analysis_space": "lognormal",
        "rescol": "res",
        "qualcol": "qual",
        "plot_marker": "o",
        "scatter_marker": "v",
        "hasData": True,
        "all_positive": True,
        "exclude": False,
        "include": True,
        "NUnique": 30,
        "bsiter": 1500,
        "N": 35,
        "ND": 7,
    }
    assert getattr(location, attr) == expected[attr]


@pytest.mark.parametrize("attr", ["fractionND", "min", "min_DL", "min_detect", "max"])
def test_locations_numbers(location, attr):
    expected = {
        "fractionND": 0.2,
        "min": 2.0,
        "min_DL": 5.0,
        "min_detect": 2.0,
        "max": 22.97,
    }
    nptest.assert_approx_equal(getattr(location, attr), expected[attr])


@pytest.mark.parametrize(
    "attr",
    [
        "useros",
        "cov",
        "geomean",
        "geostd",
        "logmean",
        "logstd",
        "mean",
        "median",
        "pctl10",
        "pctl25",
        "pctl75",
        "pctl90",
        "skew",
        "std",
    ],
)
@helpers.seed
def test_location_stats_scalars(location, attr):
    expected = {
        "useros": {True: True, False: False},
        "cov": {True: 0.5887644, False: 0.5280314},
        "geomean": {True: 8.0779865, False: 8.8140731},
        "geostd": {True: 1.8116975, False: 1.7094616},
        "logmean": {True: 2.0891426, False: 2.1763497},
        "logstd": {True: 0.5942642, False: 0.5361785},
        "mean": {True: 9.5888515, False: 10.120571},
        "median": {True: 7.5000000, False: 8.7100000},
        "pctl10": {True: 4.0460279, False: 5.0000000},
        "pctl25": {True: 5.6150000, False: 5.8050000},
        "pctl75": {True: 11.725000, False: 11.725000},
        "pctl90": {True: 19.178000, False: 19.178000},
        "skew": {True: 0.8692107, False: 0.8537566},
        "std": {True: 5.6455746, False: 5.3439797},
    }
    nptest.assert_approx_equal(
        getattr(location, attr), expected[attr][location.useros], significant=5
    )


@pytest.mark.parametrize(
    "attr",
    [
        "geomean_conf_interval",
        "logmean_conf_interval",
        "mean_conf_interval",
        "median_conf_interval",
        "shapiro",
        "shapiro_log",
        "lilliefors",
        "lilliefors_log",
        "color",
    ],
)
def test_location_stats_arrays(location, attr):
    expected = {
        "color": {
            True: [0.32157, 0.45271, 0.66667],
            False: [0.32157, 0.45271, 0.66667],
        },
        "geomean_conf_interval": {True: [6.55572, 9.79677], False: [7.25255, 10.34346]},
        "logmean_conf_interval": {True: [1.88631, 2.27656], False: [1.97075, 2.34456]},
        "mean_conf_interval": {True: [7.74564, 11.49393], False: [8.52743, 11.97627]},
        "median_conf_interval": {True: [5.66000, 8.71000], False: [6.65000, 9.850000]},
        "shapiro": {True: [0.886889, 0.001789], False: [0.896744, 0.003236]},
        "shapiro_log": {True: [0.972679, 0.520949], False: [0.964298, 0.306435]},
        "lilliefors": {True: [0.185180, 0.004200], False: [0.160353, 0.02425]},
        "lilliefors_log": {True: [0.091855, 0.64099], False: [0.08148, 0.80351]},
    }
    nptest.assert_array_almost_equal(
        getattr(location, attr), expected[attr][location.useros], decimal=5
    )


@pytest.mark.parametrize("attr", ["anderson", "anderson_log"])
@pytest.mark.parametrize("index", [0, 1, 2, pytest.param(3, marks=pytest.mark.skip), 4])
def test_location_anderson(location, attr, index):
    expected = {
        "anderson": {
            True: (
                1.54388800,
                [0.527, 0.6, 0.719, 0.839, 0.998],
                [15.0, 10.0, 5.0, 2.5, 1.0],
                None,
                0.000438139,
            ),
            False: (
                1.4392085,
                [0.527, 0.6, 0.719, 0.839, 0.998],
                [15.0, 10.0, 5.0, 2.5, 1.0],
                None,
                0.00080268,
            ),
        },
        "anderson_log": {
            True: (
                0.30409634,
                [0.527, 0.6, 0.719, 0.839, 0.998],
                [15.0, 10.0, 5.0, 2.5, 1.0],
                None,
                0.552806894,
            ),
            False: (
                0.3684061,
                [0.527, 0.6, 0.719, 0.839, 0.998],
                [15.0, 10.0, 5.0, 2.5, 1.0],
                None,
                0.41004028,
            ),
        },
    }
    result = expected[attr][location.useros][index]
    if index in [0, 4]:
        nptest.assert_approx_equal(getattr(location, attr)[index], result, significant=5)
    elif index != 3:
        nptest.assert_array_almost_equal(getattr(location, attr)[index], result, decimal=5)


@pytest.fixture
def dataset():
    known_bsiter = 750
    in_data = helpers.getTestROSData()
    in_data["res"] += 3

    out_data = helpers.getTestROSData()
    out_data["res"] -= 1.5
    influent = Location(
        in_data,
        station_type="inflow",
        bsiter=known_bsiter,
        rescol="res",
        qualcol="qual",
        useros=False,
    )

    effluent = Location(
        out_data,
        station_type="outflow",
        bsiter=known_bsiter,
        rescol="res",
        qualcol="qual",
        useros=False,
    )

    return Dataset(influent, effluent)


def test_ds_data(dataset):
    assert hasattr(dataset, "data")
    assert isinstance(dataset.data, pandas.DataFrame)


def test_paired_data(dataset):
    assert hasattr(dataset, "paired_data")
    assert isinstance(dataset.paired_data, pandas.DataFrame)


def test__non_paired_stats(dataset):
    assert hasattr(dataset, "_non_paired_stats")
    known__non_paired_stats = True
    assert dataset._non_paired_stats == known__non_paired_stats


def test__paired_stats(dataset):
    known__paired_stats = True
    assert hasattr(dataset, "_paired_stats")
    assert dataset._paired_stats == known__paired_stats


def test_name(dataset):
    assert hasattr(dataset, "name")
    assert dataset.name is None


def test_name_set(dataset):
    assert hasattr(dataset, "name")
    testname = "Test Name"
    dataset.name = testname
    assert dataset.name == testname


def test_defintion_default(dataset):
    assert hasattr(dataset, "definition")
    assert dataset.definition == {}


def test_defintion_set(dataset):
    known_definition = {"attr1": "test1", "attr2": "test2"}
    assert hasattr(dataset, "definition")
    dataset.definition = known_definition
    assert dataset.definition == known_definition


def test_include_exclude(dataset):
    known_include = True
    assert hasattr(dataset, "include")
    assert dataset.include == known_include

    assert hasattr(dataset, "exclude")
    assert dataset.exclude == (not dataset.include)


def test_wilcoxon(dataset):
    known_wilcoxon_stats = (0.0, 5.820766e-11)
    known_wilcoxon_z = known_wilcoxon_stats[0]
    known_wilcoxon_p = known_wilcoxon_stats[1]

    assert hasattr(dataset, "wilcoxon_z")
    nptest.assert_allclose(dataset.wilcoxon_z, known_wilcoxon_z, rtol=TOLERANCE)

    assert hasattr(dataset, "wilcoxon_p")
    nptest.assert_allclose(dataset.wilcoxon_p, known_wilcoxon_p, rtol=TOLERANCE)

    assert hasattr(dataset, "_wilcoxon_stats")
    nptest.assert_allclose(dataset._wilcoxon_stats, known_wilcoxon_stats, rtol=TOLERANCE)


def test_mannwhitney(dataset):
    known_mannwhitney_stats = (927.0, 2.251523e-04)
    known_mannwhitney_u = known_mannwhitney_stats[0]
    known_mannwhitney_p = known_mannwhitney_stats[1]
    assert hasattr(dataset, "mannwhitney_u")
    nptest.assert_allclose(dataset.mannwhitney_u, known_mannwhitney_u, rtol=TOLERANCE)

    assert hasattr(dataset, "mannwhitney_p")
    nptest.assert_allclose(dataset.mannwhitney_p, known_mannwhitney_p, rtol=TOLERANCE)

    assert hasattr(dataset, "_mannwhitney_stats")
    nptest.assert_allclose(
        dataset._mannwhitney_stats,
        known_mannwhitney_stats,
        rtol=TOLERANCE,
    )


@pytest.mark.xfail(OLD_SCIPY, reason="Scipy < 0.19")
def test_kendall(dataset):
    known_kendall_stats = (1.00, 5.482137e-17)
    known_kendall_tau = known_kendall_stats[0]
    known_kendall_p = known_kendall_stats[1]
    assert hasattr(dataset, "kendall_tau")
    nptest.assert_allclose(dataset.kendall_tau, known_kendall_tau, rtol=TOLERANCE)

    assert hasattr(dataset, "kendall_p")
    nptest.assert_allclose(dataset.kendall_p, known_kendall_p, rtol=TOLERANCE)

    assert hasattr(dataset, "_kendall_stats")
    nptest.assert_allclose(dataset._kendall_stats, known_kendall_stats, rtol=TOLERANCE)


def test_spearman(dataset):
    known_spearman_stats = (1.0, 0.0)
    known_spearman_rho = known_spearman_stats[0]
    known_spearman_p = known_spearman_stats[1]
    assert hasattr(dataset, "spearman_rho")
    nptest.assert_allclose(dataset.spearman_rho, known_spearman_rho, atol=0.0001)

    assert hasattr(dataset, "spearman_p")
    nptest.assert_allclose(dataset.spearman_p, known_spearman_p, atol=0.0001)

    assert hasattr(dataset, "_spearman_stats")
    nptest.assert_allclose(dataset._spearman_stats, known_spearman_stats, atol=0.0001)


def test_theil(dataset):
    known_theil_stats = (1.0, -4.5, 1.0, 1.0)
    known_theil_hislope = known_theil_stats[0]
    known_theil_intercept = known_theil_stats[1]
    known_theil_loslope = known_theil_stats[2]
    known_theil_medslope = known_theil_stats[3]
    assert hasattr(dataset, "theil_medslope")
    nptest.assert_allclose(dataset.theil_medslope, known_theil_medslope, rtol=TOLERANCE)

    assert hasattr(dataset, "theil_intercept")
    nptest.assert_allclose(dataset.theil_intercept, known_theil_intercept, rtol=TOLERANCE)

    assert hasattr(dataset, "theil_loslope")
    nptest.assert_allclose(dataset.theil_loslope, known_theil_loslope, rtol=TOLERANCE)

    assert hasattr(dataset, "theil_hislope")
    nptest.assert_allclose(dataset.theil_hislope, known_theil_hislope, rtol=TOLERANCE)

    assert hasattr(dataset, "_theil_stats")
    nptest.assert_almost_equal(dataset._theil_stats["medslope"], known_theil_stats[0], decimal=4)

    nptest.assert_almost_equal(dataset._theil_stats["intercept"], known_theil_stats[1], decimal=4)

    nptest.assert_almost_equal(dataset._theil_stats["loslope"], known_theil_stats[2], decimal=4)

    nptest.assert_almost_equal(dataset._theil_stats["hislope"], known_theil_stats[3], decimal=4)

    assert not dataset._theil_stats["is_inverted"]

    assert "estimated_effluent" in list(dataset._theil_stats.keys())
    assert "estimate_error" in list(dataset._theil_stats.keys())


def test_medianCIsOverlap(dataset):
    known_medianCIsOverlap = False
    assert known_medianCIsOverlap == dataset.medianCIsOverlap


def test__repr__normal(dataset):
    dataset.__repr__


def test_repr__None(dataset):
    dataset.definition = None
    dataset.__repr__



import difflib
import os
import re
import subprocess
import sys
from collections import namedtuple
from contextlib import contextmanager
from functools import wraps
from io import StringIO

import numpy
import pandas
import pytest
from packaging.version import Version
from pkg_resources import resource_filename


def get_img_tolerance():
    return int(os.environ.get("MPL_IMGCOMP_TOLERANCE", 15))


def seed(func):
    """Decorator to seed the RNG before any function."""

    @wraps(func)
    def wrapper(*args, **kwargs):
        numpy.random.seed(0)
        return func(*args, **kwargs)

    return wrapper


def raises(error):
    """Wrapper around pytest.raises to support None."""
    if error:
        return pytest.raises(error)
    else:

        @contextmanager
        def not_raises():
            try:
                yield
            except Exception as e:
                raise e

        return not_raises()


def requires(module, modulename):
    def outer_wrapper(function):
        @wraps(function)
        def inner_wrapper(*args, **kwargs):
            if module is None:
                raise RuntimeError(f"{modulename} required for `{function.__name__}`")
            else:
                return function(*args, **kwargs)

        return inner_wrapper

    return outer_wrapper


@seed
def make_dc_data(ndval="ND", rescol="res", qualcol="qual"):
    dl_map = {
        "A": 0.1,
        "B": 0.2,
        "C": 0.3,
        "D": 0.4,
        "E": 0.1,
        "F": 0.2,
        "G": 0.3,
        "H": 0.4,
    }

    index = pandas.MultiIndex.from_product(
        [
            list("ABCDEFGH"),
            list("1234567"),
            ["GA", "AL", "OR", "CA"],
            ["Inflow", "Outflow", "Reference"],
        ],
        names=["param", "bmp", "state", "loc"],
    )

    array = numpy.random.lognormal(mean=0.75, sigma=1.25, size=len(index))
    data = pandas.DataFrame(data=array, index=index, columns=[rescol])
    data["DL"] = data.apply(lambda r: dl_map.get(r.name[0]), axis=1)

    data[rescol] = data.apply(
        lambda r: dl_map.get(r.name[0]) if r[rescol] < r["DL"] else r[rescol], axis=1
    )

    data[qualcol] = data.apply(lambda r: ndval if r[rescol] <= r["DL"] else "=", axis=1)

    return data


@seed
def make_dc_data_complex(dropsome=True):
    dl_map = {
        "A": 0.25,
        "B": 0.50,
        "C": 0.10,
        "D": 1.00,
        "E": 0.25,
        "F": 0.50,
        "G": 0.10,
        "H": 1.00,
    }

    index = pandas.MultiIndex.from_product(
        [
            list("ABCDEFGH"),
            list("1234567"),
            ["GA", "AL", "OR", "CA"],
            ["Inflow", "Outflow", "Reference"],
        ],
        names=["param", "bmp", "state", "loc"],
    )

    xtab = (
        pandas.DataFrame(index=index, columns=["res"]).unstack(level="param").unstack(level="state")
    )

    xtab_rows = xtab.shape[0]
    for c in xtab.columns:
        mu = numpy.random.uniform(low=-1.7, high=2)
        sigma = numpy.random.uniform(low=0.1, high=2)
        xtab[c] = numpy.random.lognormal(mean=mu, sigma=sigma, size=xtab_rows)

    data = xtab.stack(level="state").stack(level="param")

    data["DL"] = data.apply(lambda r: dl_map.get(r.name[-1]), axis=1)

    data["res"] = data.apply(
        lambda r: dl_map.get(r.name[-1]) if r["res"] < r["DL"] else r["res"], axis=1
    )

    data["qual"] = data.apply(lambda r: "<" if r["res"] <= r["DL"] else "=", axis=1)

    if dropsome:
        if int(dropsome) == 1:
            dropsome = 0.25
        index = numpy.random.uniform(size=data.shape[0]) >= dropsome
        data = data.loc[index]

    return data


def comp_statfxn(x, y):
    stat = namedtuple("teststat", ("statistic", "pvalue"))
    result = x.max() - y.min()
    return stat(result, result * 0.25)


def test_data_path(filename):
    path = resource_filename("wqio.tests._data", filename)
    return path


def getTestROSData():
    """
    Generates test data for an ROS estimate.
    Input:
        None

    Output:
        Structured array with the values (results or DLs) and qualifers
        (blank or "ND" for non-detects)
    """
    raw_csv = StringIO(
        "res,qual\n2.00,=\n4.20,=\n4.62,=\n5.00,ND\n5.00,ND\n5.50,ND\n"
        "5.57,=\n5.66,=\n5.75,ND\n5.86,=\n6.65,=\n6.78,=\n6.79,=\n7.50,=\n"
        "7.50,=\n7.50,=\n8.63,=\n8.71,=\n8.99,=\n9.50,ND\n9.50,ND\n9.85,=\n"
        "10.82,=\n11.00,ND\n11.25,=\n11.25,=\n12.20,=\n14.92,=\n16.77,=\n"
        "17.81,=\n19.16,=\n19.19,=\n19.64,=\n20.18,=\n22.97,=\n"
    )

    return pandas.read_csv(raw_csv)


def compare_versions(utility="latex"):  # pragma: no cover
    "return True if a is greater than or equal to b"
    requirements = {"latex": "3.1415"}

    available = {"latex": checkdep_tex()}

    required = requirements[utility]
    present = available[utility]
    if present:
        present = Version(present)
        required = Version(required)
        return present >= required

    else:
        return False


def _show_package_info(package, name):  # pragma: no cover
    packagedir = os.path.dirname(package.__file__)
    print(f"{name} version {package.__version__} is installed in {packagedir}")


def _show_system_info():  # pragma: no cover
    import pytest

    pyversion = sys.version.replace("\n", "")
    print("Python version %s" % pyversion)
    print("pytest version %d.%d.%d" % pytest.__versioninfo__)

    import numpy

    _show_package_info(numpy, "numpy")

    import scipy

    _show_package_info(scipy, "scipy")

    import matplotlib

    _show_package_info(matplotlib, "matplotlib")

    import statsmodels

    _show_package_info(statsmodels, "statsmodels")

    import pandas

    _show_package_info(pandas, "pandas")


def checkdep_tex():  # pragma: no cover
    if sys.version_info[0] >= 3:

        def byte2str(b):
            return b.decode("ascii")

    else:  # pragma: no cover

        def byte2str(b):
            return b

    try:
        s = subprocess.Popen(["tex", "-version"], stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        line = byte2str(s.stdout.readlines()[0])
        pattern = r"3\.1\d+"
        match = re.search(pattern, line)
        v = match.group(0)
        return v
    except (IndexError, ValueError, AttributeError, OSError):
        return None


def assert_bigstring_equal(
    input_string, known_string, input_out=None, known_out=None
):  # pragma: no cover
    if input_string != known_string:
        if input_out and known_out:
            with open(input_out, "w") as fi:
                fi.write(input_string)

            with open(known_out, "w") as fo:
                fo.write(known_string)

        message = "".join(
            difflib.ndiff(input_string.splitlines(True), known_string.splitlines(True))
        )
        raise AssertionError("Multi-line strings are unequal:\n" + message)



import logging
from collections import namedtuple

import numpy
import scipy.stats as stats
from probscale.algo import _estimate_from_fit

from wqio import utils

_logger = logging.getLogger(__name__)

fitestimate = namedtuple(
    "BootstrappedFitEstimate", ["xhat", "yhat", "lower", "upper", "xlog", "ylog"]
)

__all__ = ["BCA", "percentile", "fit"]


def _acceleration(data):
    """Compute the acceleration statistic.

    Parameters
    ----------
    data : array-like
        Sequence of values whose acceleration will be computed.

    Returns
    -------
    acc : float
        The acceleration statistic.

    """

    # intermediate values
    sumcube_resids = ((data.mean() - data) ** 3).sum()

    # dodge the ZeroDivision error
    sumsqr_resids = max(((data.mean() - data) ** 2).sum(), 1e-12)

    # compute and return the acceleration
    return sumcube_resids / (6 * sumsqr_resids**1.5)


def _make_boot_index(elements, niter):
    """Generate an array of bootstrap sample sets

    Parameters
    ----------
    elements : int
        The number of rows in the original dataset.
    niter : int
        Number of iteration for the bootstrapping.

    Returns
    -------
    index : numpy array
        A collection of random *indices* that can be used to randomly
        sample a dataset ``niter`` times.

    """
    return numpy.random.randint(low=0, high=elements, size=(niter, elements))


def BCA(data, statfxn, niter=10000, alpha=0.05, log=True, warn=False):
    """
    Estimates confidence intervals around a statistic using the
    Bias-Corrected and Accelerated method.

    Parameters
    ----------
    data : array-like
        Input sequence of values
    statfxn : callable
        A reducing function that returns a single value when passed
        ``data``.
    niter : int, optional (default = 10000)
        The number of iterations for which the data will be resampled
        and the statistic recomputed.
    alpha : float, optional (default = 0.05)
        The desired confidence interval subtracted from 1. For example,
        if you want the 95% CI, use ``alpha = 0.05``.

    Returns
    -------
    CI : numpy array
        Confidence intervals around the statistic.

    Examples
    --------
    >>> import numpy
    >>> import wqio
    >>> numpy.random.seed(0)
    >>> # generate fake data and bootstrap it
    >>> data = numpy.random.normal(loc=2, scale=1.75, size=37)
    >>> wqio.bootstrap.BCA(data, numpy.mean, niter=1000, alpha=0.05)
    array([1.89528037, 3.16395883])

    """

    data = numpy.asarray(data)

    index = _make_boot_index(data.shape[0], niter)
    boot_stats = statfxn(data[index], axis=-1)
    primary_result = statfxn(data)
    boot_result = boot_stats.mean()

    # number of results below the premlinary estimate
    NumBelow = numpy.sum(boot_stats < primary_result)
    if NumBelow == 0:
        NumBelow = 0.00001

    # compute the acceleration
    a_hat = _acceleration(data)

    if NumBelow == niter:
        utils.log_or_warn(
            "All results below primary_result",
            warning=UserWarning if warn else None,
            logger=_logger.debug if log else None,
        )

    # z-stats on the % of `NumBelow` and the confidence limits
    else:
        z0 = stats.norm.ppf(NumBelow / niter)
        z = stats.norm.ppf([0.5 * alpha, 1 - (0.5 * alpha)])

        # refine the confidence limits (alphas)
        zTotal = z0 + (z0 + z) / (1 - a_hat * (z0 + z))
        new_alpha = stats.norm.cdf(zTotal) * 100.0

        # confidence intervals from the new alphas
        CI = numpy.percentile(boot_stats, new_alpha)

        # fall back to the standard percentile method if the results
        # don't make any sense
        if boot_result < CI[0] or CI[1] < boot_result:
            utils.log_or_warn(
                "Secondary result outside of CI",
                warning=UserWarning if warn else None,
                logger=_logger.debug if log else None,
            )
            CI = percentile(data, statfxn, niter, alpha=alpha)

    return CI


def percentile(data, statfxn, niter=10000, alpha=0.05):
    """
    Estimates confidence intervals around a statistic using the
    percentile method.

    Parameters
    ----------
    data : array-like
        Input sequence of values
    statfxn : callable
        A reducing function that returns a single value when passed
        ``data``.
    niter : int, optional (default = 10000)
        The number of iterations for which the data will be resampled
        and the statistic recomputed.
    alpha : float, optional (default = 0.05)
        The desired confidence interval subtracted from 1. For example,
        if you want the 95% CI, use ``alpha = 0.05``.

    Returns
    -------
    CI : numpy array
        Confidence intervals around the statistic.

    Examples
    --------
    >>> import numpy
    >>> import wqio
    >>> numpy.random.seed(0)
    >>> # generate fake data and bootstrap it
    >>> data = numpy.random.normal(loc=2, scale=1.75, size=37)
    >>> wqio.bootstrap.percentile(data, numpy.median, niter=1000, alpha=0.05)
    array([2.20960993, 3.33181602])
    """

    index = _make_boot_index(data.shape[0], niter)
    boot_stats = statfxn(data[index], axis=-1)

    # compute the `alpha/2` and `1-alpha/2` percentiles of `boot_stats`
    CI = numpy.percentile(boot_stats, [alpha * 50, 100 - (alpha * 50)], axis=0)

    return CI


def fit(x, y, fitfxn, niter=10000, alpha=0.05, xlog=False, ylog=False, **kwargs):
    """
    Perform a percentile bootstrap estimate on a linear regression.

    Parameters
    ----------
    x, y : array-like
        Predictor and response variables, respectively.
    fitfxn : callable
        The function that performs the linear regression. It must return a
        sequence (length > 3) where each element is a scalar and the first and
        second elements are the slope and intercept (in that order). The first
        two parameters fed to the function must be ``x`` and ``y`` (in that
        order).
    niter : int, optional (default = 10000)
        Number of bootstrap iterations to perform.
    alpha : float, optional (default = 0.05)
        The confidence level to estimate the response variable. For example,
        when alpha = 0.05, the 97.5th and 2.5th perctile estimates will be
        returned.
    xlog, ylog : bool, optional (default = False)
        Toggles performing the fit and estimates in arithmetic or logarithmic
        space.

    Additional Parameters
    ---------------------
    Any additional keyword arguments will be directly passed to ``fitfxn``
    after ``x`` and ``y``.

    Returns
    -------
    BootstrappedFitEstimate : namedtuple of numpy.arrays
        Names include:

          xhat : the synthetic predictor variable
          yhat : the estimated response variable
          lower, upper : the confidence intervals around the estimated respone

    Examples
    --------
    >>> import numpy
    >>> from matplotlib import pyplot
    >>> from wqio import bootstrap
    >>> N = 10
    >>> x = numpy.arange(1, N + 1, dtype=float)
    >>> y = numpy.array([4.527, 3.519, 9.653, 8.036, 10.805,
    ...                  14.329, 13.508, 11.822, 13.281, 10.410])
    >>> bsfit = bootstrap.fit(x, y, numpy.polyfit, niter=2000,
    ...                       xlog=False, ylog=False,
    ...                       deg=1, full=False)
    >>> fig, ax = pyplot.subplots()
    >>> _pnts = ax.plot(x, y, 'k.', zorder=2)
    >>> _line = ax.plot(bsfit.xhat, bsfit.yhat, 'b-', zorder=1)
    >>> _ci = ax.fill_between(bsfit.xhat, bsfit.lower, bsfit.upper,
    ...                       zorder=0, alpha=0.5)

    """

    # get the data in order according to the *predictor* variable only
    sort_index = numpy.argsort(x)
    x = x[sort_index]
    y = y[sort_index]

    # apply logs if necessary
    if xlog:
        x = numpy.log(x)
    if ylog:
        y = numpy.log(y)

    # compute fit on original data
    main_params = fitfxn(x, y, **kwargs)

    # raw loop to estimate the bootstrapped fit parameters
    bs_params = numpy.array(
        [fitfxn(x[ii], y[ii], **kwargs) for ii in _make_boot_index(len(x), niter)]
    )

    # un-log, if necesssary
    if xlog:
        x = numpy.exp(x)

    # compute estimate from original data fit
    yhat = _estimate_from_fit(x, main_params[0], main_params[1], xlog=xlog, ylog=ylog)

    # full array of estimates
    bs_estimates = _estimate_from_fit(
        x[:, None], bs_params[:, 0], bs_params[:, 1], xlog=xlog, ylog=ylog
    )

    # both alpha alphas
    percentiles = 100 * numpy.array([alpha * 0.5, 1 - alpha * 0.5])

    # lower, upper bounds
    bounds = numpy.percentile(bs_estimates, percentiles, axis=1)

    return fitestimate(x, yhat, bounds[0], bounds[1], xlog, ylog)



from functools import partial

import numpy
import probscale
import seaborn
from matplotlib import pyplot, ticker
from pandas.api.types import CategoricalDtype

from wqio import utils, validate


def rotate_tick_labels(ax, rotation, which, rotation_mode="anchor", ha="right"):
    """Rotates the ticklabels of a matplotlib Axes

    Parameters
    ----------
    ax : matplotlib Axes
        The Axes object that will be modified.
    rotation : float
        The amount of rotation, in degrees, to be applied to the labels.
    which : string
        The axis whose ticklabels will be rotated. Valid values are 'x',
        'y', or 'both'.
    rotation_mode : string, optional
        The rotation point for the ticklabels. Highly recommended to use
        the default value ('anchor').
    ha : string
        The horizontal alignment of the ticks. Again, recommended to use
        the default ('right').

    Returns
    -------
    None

    """

    if which == "both":
        rotate_tick_labels(ax, rotation, "x", rotation_mode=rotation_mode, ha=ha)
        rotate_tick_labels(ax, rotation, "y", rotation_mode=rotation_mode, ha=ha)
    else:
        if which == "x":
            axis = ax.xaxis

        elif which == "y":
            axis = ax.yaxis

        for t in axis.get_ticklabels():
            t.set_horizontalalignment(ha)
            t.set_rotation(rotation)
            t.set_rotation_mode(rotation_mode)


def log_formatter(use_1x=True, threshold=5):
    def _formatter(tick, pos=None, use_1x=True, threshold=3):
        """Formats log axes as `1 x 10^N` when N > 4 or N < -4."""

        if 10**threshold >= tick > 1:
            tick = f"{int(tick):,d}"
        elif tick > 10**threshold or tick < 10 ** (-1 * threshold):
            if use_1x:
                tick = r"$1 \times 10 ^ {%d}$" % int(numpy.log10(tick))
            else:
                tick = r"$10 ^ {%d}$" % int(numpy.log10(tick))
        else:
            tick = utils.sig_figs(tick, n=1)

        return str(tick)

    func = partial(_formatter, use_1x=use_1x, threshold=threshold)
    return ticker.FuncFormatter(func)


def gridlines(ax, xlabel=None, ylabel=None, xscale=None, yscale=None, xminor=True, yminor=True):
    """Standard formatting for gridlines on a matplotlib Axes

    Parameters
    ----------
    ax : matplotlib Axes
        The Axes object that will be modified.
    xlabel, ylabel : string, optional
        The labels of the x- and y-axis.
    xscale, yscale : string, optional
        The scale of each axis. Can be 'linear', 'log', or 'prob'.
    xminor, yminor : bool, optional
        Toggles the grid on minor ticks. Has no effect if minor ticks
        are not present.

    Returns
    -------
    None

    """

    # set the scales
    if xscale is not None:
        ax.set_xscale(xscale)

    if yscale is not None:
        ax.set_yscale(yscale)

    # set the labels
    if xlabel is not None:
        ax.set_xlabel(xlabel)

    if ylabel is not None:
        ax.set_ylabel(ylabel)

    # major grids
    ax.yaxis.grid(True, which="major", ls="-", alpha=0.35)
    ax.xaxis.grid(True, which="major", ls="-", alpha=0.35)

    # minor grids
    if xminor:
        ax.xaxis.grid(True, which="minor", ls="-", alpha=0.17)

    if yminor:
        ax.yaxis.grid(True, which="minor", ls="-", alpha=0.17)


def one2one_line(ax, set_limits=True, set_aspect=True, **kwargs):
    label = kwargs.pop("label", "1:1 Line")
    axis_limits = [
        numpy.min([ax.get_xlim(), ax.get_ylim()]),
        numpy.max([ax.get_xlim(), ax.get_ylim()]),
    ]
    if set_limits:
        ax.set_xlim(axis_limits)
        ax.set_ylim(axis_limits)
    if set_aspect:
        ax.set_aspect("equal")

    return ax.plot(axis_limits, axis_limits, label=label, **kwargs)


def jointplot(
    x=None,
    y=None,
    data=None,
    xlabel=None,
    ylabel=None,
    color=None,
    zeromin=True,
    one2one=True,
):
    """Plots the joint distribution of two variables via seaborn

    Parameters
    ----------
    x, y : array-like or string
        Sequences of values or column names found within ``data``.
    data : pandas DataFrame or None, optional
        An optional DataFrame containing the data.
    xlabel, ylabel : string, optional
        Overrides the default x- and y-axis labels.
    color : matplotlib color, optional
        Color used for the plot elements.
    zeromin : bool, optional
        When True (default), force lower axes limits to 0.
    one2one : bool, optional
        When True (default), plots the 1:1 line on the axis and sets
        the x- and y-axis limits to be equal.

    Returns
    -------
    jg : seaborn.JointGrid

    """
    jg = seaborn.jointplot(x=x, y=y, color=color, data=data, marginal_kws=dict(rug=True, kde=True))

    if xlabel is None:
        xlabel = jg.ax_joint.get_xlabel()

    if ylabel is None:
        ylabel = jg.ax_joint.get_ylabel()

    jg.set_axis_labels(xlabel=xlabel, ylabel=ylabel)

    if zeromin:
        jg.ax_joint.set_xlim(left=0)
        jg.ax_joint.set_ylim(bottom=0)

    if one2one:
        ax_limit_min = numpy.min([jg.ax_joint.get_xlim(), jg.ax_joint.get_ylim()])
        ax_limit_max = numpy.max([jg.ax_joint.get_xlim(), jg.ax_joint.get_ylim()])
        jg.ax_joint.set_xlim(left=ax_limit_min, right=ax_limit_max)
        jg.ax_joint.set_ylim(bottom=ax_limit_min, top=ax_limit_max)
        jg.ax_joint.plot(
            [ax_limit_min, ax_limit_max],
            [ax_limit_min, ax_limit_max],
            marker="None",
            linestyle="-",
            linewidth=1.75,
            color=color or "k",
            alpha=0.45,
            label="1:1 line",
        )

        jg.ax_joint.legend(frameon=False, loc="upper left")

    return jg


def whiskers_and_fliers(x, q1=None, q3=None, transformout=None):
    """Computes extent of whiskers and fliers on optionally transformed
    data for box and whisker plots.

    Parameters
    ----------
    x : array-like
        Sequence of optionally transformed data.
    q1, q3 : floats, optional
        First and third quartiles of the optionally transformed data.
    transformout : callable, optional
        Function to un-transform the results back into the original
        space of the data.

    Returns
    -------
    whiskers_and_fliers : dict
        Dictionary of whisker and fliers values.

    Examples
    --------
    >>> x = numpy.random.lognormal(size=37)
    >>> whisk_fly = whiskers_and_fliers(numpy.log(x), transformout=numpy.exp)

    See also
    --------
    wqio.utils.figutils.boxplot

    """
    wnf = {}
    if transformout is None:

        def transformout(x):
            return x

    if q1 is None:
        q1 = numpy.percentile(x, 25)

    if q3 is None:
        q3 = numpy.percentile(x, 75)

    iqr = q3 - q1
    # get low extreme
    loval = q1 - (1.5 * iqr)
    whislo = numpy.compress(x >= loval, x)
    whislo = q1 if len(whislo) == 0 or numpy.min(whislo) > q1 else numpy.min(whislo)

    # get high extreme
    hival = q3 + (1.5 * iqr)
    whishi = numpy.compress(x <= hival, x)
    whishi = q3 if len(whishi) == 0 or numpy.max(whishi) < q3 else numpy.max(whishi)

    wnf["fliers"] = numpy.hstack(
        [
            transformout(numpy.compress(x < whislo, x)),
            transformout(numpy.compress(x > whishi, x)),
        ]
    )
    wnf["whishi"] = transformout(whishi)
    wnf["whislo"] = transformout(whislo)

    return wnf


def boxplot(
    boxplot_stats,
    ax=None,
    position=None,
    width=0.8,
    shownotches=True,
    color="b",
    marker="o",
    patch_artist=True,
    showmean=False,
):
    """
    Draws a boxplot on an axes

    Parameters
    ----------
    boxplot_stats : list of dicts
        List of matplotlib boxplot-compatible statistics to be plotted
    ax : matplotlib Axes, optional
        The axis on which the boxplot will be drawn.
    position : int or list of int, optional
        Location on the x-axis where the boxplot will be drawn.
    width : float, optional (default = 0.8)
        Width of the boxplots.
    shownotches : bool, optional (default = True)
        Toggles notched boxplots where the notches show a confidence
        interval around the mediand.
    color : string, optional (default = 'b')
        Matplotlib color used to plot the outliers, median or box, and
        the optional mean.
    marker : str, optional (default = 'o')
        Matplotlib marker used for the the outliers and optional mean.
    patch_artist : bool, optional (default = True)
        Toggles drawing the boxes as a patch filled in with ``color``
        and a black median or as a black where the median is drawn
        in the ``color``.
    showmean : bool, optional (default = False)
        Toggles inclusion of the means in the boxplots.

    Returns
    -------
    bp : dictionary of matplotlib artists
        The graphical elements of the boxplot.

    """

    fig, ax = validate.axes(ax)

    if position is None:
        position = numpy.arange(len(boxplot_stats)) + 1
    elif numpy.isscalar(position):
        position = [position]

    meanprops = dict(marker=marker, markersize=6, markerfacecolor=color, markeredgecolor="Black")

    flierprops = dict(
        marker=marker,
        markersize=4,
        zorder=4,
        markerfacecolor="none",
        markeredgecolor=color,
        alpha=1,
    )

    whiskerprops = dict(linestyle="-", color="k", linewidth=0.75, zorder=4)

    if patch_artist:
        medianprops = dict(linewidth=1.00, color="k", linestyle="-", zorder=5)

        boxprops = dict(edgecolor="k", facecolor=color, linewidth=0.75, zorder=4, alpha=0.5)
    else:
        medianprops = dict(linewidth=1.00, color=color, linestyle="-", zorder=3)

        boxprops = dict(color="k", linewidth=0.75, zorder=4)

    bp = ax.bxp(
        boxplot_stats,
        positions=position,
        widths=width,
        showmeans=showmean,
        meanprops=meanprops,
        flierprops=flierprops,
        whiskerprops=whiskerprops,
        medianprops=medianprops,
        boxprops=boxprops,
        shownotches=shownotches,
        showcaps=False,
        manage_ticks=False,
        patch_artist=patch_artist,
    )

    return bp


def probplot(
    data,
    ax=None,
    axtype="prob",
    yscale="log",
    xlabel=None,
    ylabel=None,
    bestfit=False,
    scatter_kws=None,
    line_kws=None,
    return_results=False,
):
    """Probability, percentile, and quantile plots.

    Parameters
    ----------
    data : sequence or array-like
        1-dimensional data to be plotted
    ax : optional matplotlib axes object or None (default).
        The Axes on which to plot. If None is provided, one will be
        created.
    axtype : string (default = 'pp')
        Type of plot to be created. Options are:
            - 'prob': probabilty plot
            - 'pp': percentile plot
            - 'qq': quantile plot
    yscale : string (default = 'log')
        Scale for the y-axis. Use 'log' for logarithmic (default) or
        'linear'.
    xlabel, ylabel : string or None (default)
        Axis labels for the plot.
    bestfit : bool, optional (default is False)
        Specifies whether a best-fit line should be added to the
        plot.
    scatter_kws, line_kws : dictionary
        Dictionary of keyword arguments passed directly to `pyplot.plot`
        when drawing the scatter points and best-fit line, respectively.
    return_results : bool (default = False)
        If True a dictionary of results of is returned along with the
        figure. Keys are:
            q - array of quantiles
            x, y - arrays of data passed to function
            xhat, yhat - arrays of modeled data plotted in best-fit line
            res - a statsmodels Result object.

    Returns
    -------
    fig : matplotlib.Figure
    result : dictionary of linear fit results.

    """
    output = probscale.viz.probplot(
        data,
        ax=ax,
        plottype=axtype,
        probax="x",
        datalabel=ylabel,
        problabel=xlabel,
        datascale=yscale,
        scatter_kws=scatter_kws,
        line_kws=line_kws,
        bestfit=bestfit,
        return_best_fit_results=return_results,
    )
    return output


def _connect_spines(left_ax, right_ax, left_y, right_y, linestyle="solid", **line_kwds):
    """Connects the y-spines between two Axes

    Parameters
    ----------
    left_ax, right_ax : matplotlib Axes objects
        The Axes that need to be connected.
    left_y, right_y : float
        Values on the spines that wil be connected.
    linestyle : string, optional (default = 'solid')
        The line style to use. Valid values are 'solid', 'dashed',
        'dashdot', 'dotted'.
    **line_kwds : keyword arguments
        Additional options for style the line.

    Returns
    -------
    connector : BboxConnector
        The weird mpl-line-like-thingy that connects the spines.

    """

    import matplotlib.transforms as mtrans
    import mpl_toolkits.axes_grid1.inset_locator as inset

    left_trans = mtrans.blended_transform_factory(left_ax.transData, left_ax.transAxes)
    right_trans = mtrans.blended_transform_factory(right_ax.transData, right_ax.transAxes)

    left_data_trans = left_ax.transScale + left_ax.transLimits
    right_data_trans = right_ax.transScale + right_ax.transLimits

    left_pos = left_data_trans.transform((0, left_y))[1]
    right_pos = right_data_trans.transform((0, right_y))[1]

    bbox = mtrans.Bbox.from_extents(0, left_pos, 0, right_pos)
    right_bbox = mtrans.TransformedBbox(bbox, right_trans)
    left_bbox = mtrans.TransformedBbox(bbox, left_trans)

    # deal with the linestyle
    connector = inset.BboxConnector(
        left_bbox, right_bbox, loc1=3, loc2=2, linestyle=linestyle, **line_kwds
    )
    connector.set_clip_on(False)
    left_ax.add_artist(connector)

    return connector


def parallel_coordinates(dataframe, hue, cols=None, palette=None, showlegend=True, **subplot_kws):
    """Produce a parallel coordinates plot from a dataframe.

    Parameters
    ----------
    dataframe : pandas.DataFrame
        The data to be plotted.
    hue : string
        The column used to the determine assign the lines' colors.
    cols : list of strings, optional
        The non-hue columns to include. If None, all other columns are
        used.
    palette : string, optional
        Name of the seaborn color palette to use.
    showlegend : bool (default = True)
        Toggles including a legend on the plot.
    **subplot_kws : keyword arguments
        Options passed directly to pyplot.subplots()

    Returns
    -------
    fig : matplotlib Figure

    """

    # get the (non-hue) columns to plot
    if cols is None:
        cols = dataframe.columns.tolist()
        cols.remove(hue)

    # subset the data, putting the hue column last
    # python 3.5: data = dataframe[[*cols, hue]]
    data = dataframe[[*cols, hue]]

    # these plots look ridiculous in anything other than 'ticks'
    with seaborn.axes_style("ticks"):
        fig, axes = pyplot.subplots(ncols=len(cols), **subplot_kws)
        hue_vals = dataframe[hue].unique()
        colors = seaborn.color_palette(palette=palette, n_colors=len(hue_vals))
        color_dict = {}
        lines = []
        for h, c in zip(hue_vals, colors):
            lines.append(pyplot.Line2D([0], [0], linestyle="-", color=c, label=h))
            color_dict[h] = c

        for col, ax in zip(cols, axes):
            data_limits = [(0, dataframe[col].min()), (0, dataframe[col].max())]
            ax.set_xticks([0])
            ax.update_datalim(data_limits)
            ax.set_xticklabels([col])
            ax.autoscale(axis="y")
            ax.tick_params(axis="y", direction="inout")
            ax.tick_params(axis="x", direction="in")

        for row in data.values:
            for n, (ax1, ax2) in enumerate(zip(axes[:-1], axes[1:])):
                _connect_spines(ax1, ax2, row[n], row[n + 1], color=color_dict[row[-1]])

    if showlegend:
        fig.legend(lines, hue_vals)

    fig.subplots_adjust(wspace=0)
    seaborn.despine(fig=fig, bottom=True, trim=True)
    return fig


def categorical_histogram(df, valuecol, bins, classifier=None, **factoropts):
    """Plot a faceted, categorical histogram.

    Parameters
    ----------
    df : pandas.DataFrame
        Dataframe of storm information such as precipitation depth,
        duration, presence of outflow, flow volume, etc.
    valuecol : str
        The name of the column that should be categorized and plotted.
    bins : array-like
        The right-edges of the histogram bins.
    classifier : callable, optional
        Function-like object that classifies the values in ``valuecol``.
        Should accept a float scalar and return a string.
    factoropts : keyword arguments, optional
        Options passed directly to seaborn.factorplot

    Returns
    -------
    fig : seaborn.FacetGrid

    See also
    --------
    seaborn.factorplot

    """

    def format_col(colname):
        return colname.replace("_", " ").title()

    def process_column(colname):
        if colname is not None:
            return format_col(colname)

    if classifier is None:
        classifier = partial(utils.classifier, bins=bins, units="mm")

    cats = utils.unique_categories(classifier, bins)
    cat_type = CategoricalDtype(cats, ordered=True)
    aspect = factoropts.pop("aspect", 1.6)

    display_col = format_col(valuecol)

    processed_opts = dict(
        row=process_column(factoropts.pop("row", None)),
        col=process_column(factoropts.pop("col", None)),
        hue=process_column(factoropts.pop("hue", None)),
        kind="count",
        aspect=aspect,
        sharex=True,
    )

    final_opts = {**factoropts, **processed_opts}

    fig = (
        df.assign(display=df[valuecol].apply(classifier).astype(cat_type))
        .drop([valuecol], axis=1)
        .rename(columns={"display": valuecol})
        .rename(columns=lambda c: format_col(c))
        .pipe((seaborn.catplot, "data"), x=display_col, **final_opts)
        .set_ylabels("Occurences")
    )

    return fig
